/*-
 * Copyright (c) 2004 - 2010 CTPP Team
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the CTPP Team nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      CDT.hpp
 *
 * $CTPP$
 */
#ifndef _CDT_HPP__
#define _CDT_HPP__ 1

/**
  @file CDT.hpp
  @brief Common Data Type
*/

#include <map>
#include <string>
#include <vector>

#include "CTPP2Exception.hpp"

namespace CTPP // C++ Template Engine
{

#define C_MAX_SPRINTF_LENGTH 128

/**
  @class CDT CDT.hpp <CDT.hpp>
  @brief Common Data Type
*/
class CTPP2DECL CDT
{
private:
	/**
	  @var typedef std::vector<CDT> Vector
	  @brief internal array definition
	*/
	typedef std::vector<CDT>             Vector;

	/**
	  @var typedef std::map<std::string, CDT> Map
	  @brief internal hash definition
	*/
	typedef std::map<std::string, CDT>   Map;
public:
	/**
	  @enum eValType CDT.hpp <CDT.hpp>
	  @brief Describes type of stored value
	*/
	enum eValType { UNDEF = 0x01, INT_VAL = 0x02, REAL_VAL = 0x03, POINTER_VAL = 0x04, STRING_VAL = 0x15, ARRAY_VAL = 0x16, HASH_VAL = 0x17};

	/**
	  @brief Constructor
	  @param oValue - type of value
	*/
	CDT(const CDT::eValType & oValue = UNDEF);

	/**
	  @brief Copy constructor
	  @param oCDT - Object to copy
	*/
	CDT(const CDT & oCDT);

	/**
	  @brief Copy operator
	  @param oCDT - Object to copy
	  @return read/write referense to self
	*/
	CDT & operator=(const CDT & oCDT);

	/**
	  @brief Type cast constructor
	  @param oValue - INT_64 value
	*/
	CDT(const INT_64 & oValue);

	/**
	  @brief Type cast constructor
	  @param oValue - UINT_64 value
	*/
	CDT(const UINT_64 & oValue);

	/**
	  @brief Type cast constructor
	  @param oValue - INT_32 value
	*/
	CDT(const INT_32 & oValue);

	/**
	  @brief Type cast constructor
	  @param oValue - UINT_32 value
	*/
	CDT(const UINT_32 & oValue);

	/**
	  @brief Type cast constructor
	  @param oValue - W_FLOAT value
	*/
	CDT(const W_FLOAT & oValue);

	/**
	  @brief Type cast constructor
	  @param oValue - string value
	*/
	CDT(const std::string & oValue);

	/**
	  @brief Type cast constructor
	  @param oValue - asciz string to copy
	  @return read/write referense to self
	*/
	CDT(CCHAR_P oValue);

	/**
	  @brief Type cast constructor
	  @param oValue - generic pointer value
	*/
	CDT(void * oValue);

	/**
	  @brief Copy operator
	  @param oValue - INT_64 value to copy
	  @return read/write referense to self
	*/
	CDT & operator=(const INT_64 & oValue);

	/**
	  @brief Copy operator
	  @param oValue - UINT_64 value to copy
	  @return read/write referense to self
	*/
	CDT & operator=(const UINT_64 & oValue);

	/**
	  @brief Copy operator
	  @param oValue - INT_32 value to copy
	  @return read/write referense to self
	*/
	CDT & operator=(const INT_32 & oValue);

	/**
	  @brief Copy operator
	  @param oValue - UINT_32 value to copy
	  @return read/write referense to self
	*/
	CDT & operator=(const UINT_32 & oValue);

	/**
	  @brief Copy operator
	  @param oValue - W_FLOAT value to copy
	  @return read/write referense to self
	*/
	CDT & operator=(const W_FLOAT & oValue);

	/**
	  @brief Copy operator
	  @param oValue - string to copy
	  @return read/write referense to self
	*/
	CDT & operator=(const std::string & oValue);

	/**
	  @brief Copy operator
	  @param oValue - asciz string to copy
	  @return read/write referense to self
	*/
	CDT & operator=(CCHAR_P oValue);

	/**
	  @brief Copy operator
	  @param oValue - generic pointer
	  @return read/write referense to self
	*/
	CDT & operator=(void * oValue);

	/**
	  @brief Provides access to the data contained in CDT
	  @param iPos - The index of the element
	  @return Object with data
	*/
	CDT & operator[](const UINT_32 & iPos);

	/**
	  @brief Provides access to the data contained in CDT
	  @param sKey - The key of the element
	  @return Object with data
	*/
	CDT & operator[](const std::string & sKey);

	/**
	  @brief Provides constant access to the data contained in CDT
	  @param iPos - The index of the element
	  @return Object with data
	*/
	CDT GetCDT(const UINT_32 & iPos) const;

	/**
	  @brief Provides constant access to the data contained in CDT
	  @param sKey - The key of the hash
	  @return Object with data
	*/
	CDT GetCDT(const std::string & sKey) const;

	/**
	  @brief Provides constant access to the data contained in CDT
	  @param sKey - The key of the hash [in]
	  @param iCDTExist - Existence flag [out], is set to 0 if object exist or -1 otherwise
	  @return Object with data
	*/
	CDT GetExistedCDT(const std::string & sKey, INT_32 & iCDTExist) const;

	/**
	  @brief Erase element from HASH
	  @param sKey - The key of the hash [in]
	  @return true - if key found, false - otherwise
	*/
	bool Erase(const std::string & sKey);

	/**
	  @brief Check element in hash
	  @param sKey - The key of the hash
	  @return true, if element present, false - otherwise
	*/
	bool Exists(const std::string & sKey) const;

	/**
	  @brief Check element in array
	  @param iPos - The index of the element
	  @return true, if element present, false - otherwise
	*/
	bool Exists(const UINT_32 & iPos) const;

	/**
	  @brief Push value into array
	  @param oValue - INT_64 value
	*/
	void PushBack(const INT_64 & oValue);

	/**
	  @brief Push value into array
	  @param oValue - UINT_64 value
	*/
	void PushBack(const UINT_64 & oValue);

	/**
	  @brief Push value into array
	  @param oValue - INT_32 value
	*/
	void PushBack(const INT_32 & oValue);

	/**
	  @brief Push value into array
	  @param oValue - UINT_32 value
	*/
	void PushBack(const UINT_32 & oValue);

	/**
	  @brief Push value into array
	  @param oValue - W_FLOAT value
	*/
	void PushBack(const W_FLOAT & oValue);

	/**
	  @brief Push value into array
	  @param oValue - string value
	*/
	void PushBack(const std::string & oValue);

	/**
	  @brief Push value into array
	  @param oValue - asciz string
	*/
	void PushBack(CCHAR_P oValue);

	/**
	  @brief Push value into array
	  @param oValue - CDT object
	*/
	void PushBack(const CDT & oValue);

	/**
	  @brief Returns a boolean value telling whether object has a value
	  @return true if object has a value, false - otherwise
	*/
	bool Defined();

	/**
	  @brief Provides range-check access to the data contained in CDT
	  @param iPos - The index of the element
	  @return Read/write reference to data
	*/
	CDT & At(const UINT_32 & iPos);

	/**
	  @brief Provides range-check access to the data contained in CDT
	  @param sKey - The key of the hash
	  @return Read/write reference to data
	*/
	CDT & At(const std::string & sKey);

	// Operator + ////////////////////////////////////

	/**
	  @brief Operator + for INT_64 argument
	  @param oValue - value to add
	  @return new CDT object
	*/
	CDT operator+(const INT_64 & oValue);

	/**
	  @brief Operator + for UINT_64 argument
	  @param oValue - value to add
	  @return new CDT object
	*/
	CDT operator+(const UINT_64 & oValue);

	/**
	  @brief Operator + for INT_32 argument
	  @param oValue - value to add
	  @return new CDT object
	*/
	CDT operator+(const INT_32 & oValue);

	/**
	  @brief Operator + for UINT_32 argument
	  @param oValue - value to add
	  @return new CDT object
	*/
	CDT operator+(const UINT_32 & oValue);

	/**
	  @brief Operator + for W_FLOAT argument
	  @param oValue - value to add
	  @return new CDT object
	*/
	CDT operator+(const W_FLOAT & oValue);

#if(0)
	/**
	  @brief Operator + for string argument
	  @param oValue - value to add
	  @return new CDT object
	*/
	CDT operator+(const std::string & oValue);
#endif
	/**
	  @brief Operator + for CDT argument
	  @param oCDT - value to add
	  @return new CDT object
	*/
	CDT operator+(const CDT & oCDT);

	// Operator - ////////////////////////////////////

	/**
	  @brief Operator - for INT_64 argument
	  @param oValue - value to substract
	  @return new CDT object
	*/
	CDT operator-(const INT_64 & oValue);

	/**
	  @brief Operator - for UINT_64 argument
	  @param oValue - value to substract
	  @return new CDT object
	*/
	CDT operator-(const UINT_64 & oValue);

	/**
	  @brief Operator - for INT_32 argument
	  @param oValue - value to substract
	  @return new CDT object
	*/
	CDT operator-(const INT_32 & oValue);

	/**
	  @brief Operator - for UINT_32 argument
	  @param oValue - value to substract
	  @return new CDT object
	*/
	CDT operator-(const UINT_32 & oValue);

	/**
	  @brief Operator - for W_FLOAT argument
	  @param oValue - value to substract
	  @return new CDT object
	*/
	CDT operator-(const W_FLOAT & oValue);

	/**
	  @brief Operator - for CDT argument
	  @param oCDT - value to substract
	  @return new CDT object
	*/
	CDT operator-(const CDT & oCDT);

	// Operator * ////////////////////////////////////

	/**
	  @brief Operator * for INT_64 argument
	  @param oValue - value to multiplicate
	  @return new CDT object
	*/
	CDT operator*(const INT_64 & oValue);

	/**
	  @brief Operator * for UINT_64 argument
	  @param oValue - value to multiplicate
	  @return new CDT object
	*/
	CDT operator*(const UINT_64 & oValue);

	/**
	  @brief Operator * for INT_32 argument
	  @param oValue - value to multiplicate
	  @return new CDT object
	*/
	CDT operator*(const INT_32 & oValue);

	/**
	  @brief Operator * for UINT_32 argument
	  @param oValue - value to multiplicate
	  @return new CDT object
	*/
	CDT operator*(const UINT_32 & oValue);

	/**
	  @brief Operator * for W_FLOAT argument
	  @param oValue - value to multiplicate
	  @return new CDT object
	*/
	CDT operator*(const W_FLOAT & oValue);

	/**
	  @brief Operator * for CDT argument
	  @param oCDT - value to multiplicate
	  @return new CDT object
	*/
	CDT operator*(const CDT & oCDT);

	// Operator / ////////////////////////////////////

	/**
	  @brief Operator / for INT_64 argument
	  @param oValue - value to divide
	  @return new CDT object
	*/
	CDT operator/(const INT_64 & oValue);

	/**
	  @brief Operator / for UINT_64 argument
	  @param oValue - value to divide
	  @return new CDT object
	*/
	CDT operator/(const UINT_64 & oValue);

	/**
	  @brief Operator / for INT_32 argument
	  @param oValue - value to divide
	  @return new CDT object
	*/
	CDT operator/(const INT_32 & oValue);

	/**
	  @brief Operator / for UINT_32 argument
	  @param oValue - value to divide
	  @return new CDT object
	*/
	CDT operator/(const UINT_32 & oValue);

	/**
	  @brief Operator / for W_FLOAT argument
	  @param oValue - value to divide
	  @return new CDT object
	*/
	CDT operator/(const W_FLOAT & oValue);

	/**
	  @brief Operator / for CDT argument
	  @param oCDT - value to divide
	  @return new CDT object
	*/
	CDT operator/(const CDT & oCDT);

	// ///////////////////////////////////////////////

	// Operator += ///////////////////////////////////

	/**
	  @brief Operator += for INT_64 argument
	  @param oValue - value to add
	  @return Read/write reference to self
	*/
	CDT & operator+=(const INT_64 & oValue);

	/**
	  @brief Operator += for UINT_64 argument
	  @param oValue - value to add
	  @return Read/write reference to self
	*/
	CDT & operator+=(const UINT_64 & oValue);

	/**
	  @brief Operator += for INT_32 argument
	  @param oValue - value to add
	  @return Read/write reference to self
	*/
	CDT & operator+=(const INT_32 & oValue);

	/**
	  @brief Operator += for UINT_32 argument
	  @param oValue - value to divide
	  @return Read/write reference to self
	*/
	CDT & operator+=(const UINT_32 & oValue);

	/**
	  @brief Operator += for W_FLOAT argument
	  @param oValue - value to add
	  @return Read/write reference to self
	*/
	CDT & operator+=(const W_FLOAT & oValue);

	/**
	  @brief Operator += for string argument
	  @param oValue - value to add
	  @return Read/write reference to self
	*/
	CDT & operator+=(const std::string & oValue);

	/**
	  @brief Operator += for CDT argument
	  @param oCDT - value to add
	  @return Read/write reference to self
	*/
	CDT & operator+=(const CDT & oCDT);

	// Operator -= ///////////////////////////////////

	/**
	  @brief Operator -= for INT_64 argument
	  @param oValue - value to substract
	  @return Read/write reference to self
	*/
	CDT & operator-=(const INT_64 & oValue);

	/**
	  @brief Operator -= for UINT_64 argument
	  @param oValue - value to substract
	  @return Read/write reference to self
	*/
	CDT & operator-=(const UINT_64 & oValue);

	/**
	  @brief Operator -= for INT_32 argument
	  @param oValue - value to substract
	  @return Read/write reference to self
	*/
	CDT & operator-=(const INT_32 & oValue);

	/**
	  @brief Operator -= for UINT_32 argument
	  @param oValue - value to substract
	  @return Read/write reference to self
	*/
	CDT & operator-=(const UINT_32 & oValue);

	/**
	  @brief Operator -= for W_FLOAT argument
	  @param oValue - value to substract
	  @return Read/write reference to self
	*/
	CDT & operator-=(const W_FLOAT & oValue);

	/**
	  @brief Operator -= for string argument
	  @param oCDT - value to substract
	  @return Read/write reference to self
	*/
	CDT & operator-=(const CDT & oCDT);

	// Operator *= ///////////////////////////////////

	/**
	  @brief Operator *= for INT_64 argument
	  @param oValue - value to multiplication
	  @return Read/write reference to self
	*/
	CDT & operator*=(const INT_64 & oValue);

	/**
	  @brief Operator *= for UINT_64 argument
	  @param oValue - value to multiplication
	  @return Read/write reference to self
	*/
	CDT & operator*=(const UINT_64 & oValue);

	/**
	  @brief Operator *= for INT_32 argument
	  @param oValue - value to multiplication
	  @return Read/write reference to self
	*/
	CDT & operator*=(const INT_32 & oValue);

	/**
	  @brief Operator *= for UINT_32 argument
	  @param oValue - value to multiplication
	  @return Read/write reference to self
	*/
	CDT & operator*=(const UINT_32 & oValue);

	/**
	  @brief Operator *= for W_FLOAT argument
	  @param oValue - value to multiplication
	  @return Read/write reference to self
	*/
	CDT & operator*=(const W_FLOAT & oValue);

	/**
	  @brief Operator *= for CDT argument
	  @param oCDT - value to multiplication
	  @return Read/write reference to self
	*/
	CDT & operator*=(const CDT & oCDT);

	// Operator /= ///////////////////////////////////

	/**
	  @brief Operator /= for INT_64 argument
	  @param oValue - value to division
	  @return Read/write reference to self
	*/
	CDT & operator/=(const INT_64 & oValue);

	/**
	  @brief Operator /= for UINT_64 argument
	  @param oValue - value to division
	  @return Read/write reference to self
	*/
	CDT & operator/=(const UINT_64 & oValue);

	/**
	  @brief Operator /= for INT_32 argument
	  @param oValue - value to division
	  @return Read/write reference to self
	*/
	CDT & operator/=(const INT_32 & oValue);

	/**
	  @brief Operator /= for UINT_32 argument
	  @param oValue - value to division
	  @return Read/write reference to self
	*/
	CDT & operator/=(const UINT_32 & oValue);

	/**
	  @brief Operator /= for W_FLOAT argument
	  @param oValue - value to division
	  @return Read/write reference to self
	*/
	CDT & operator/=(const W_FLOAT & oValue);

	/**
	  @brief Operator /= for CDT argument
	  @param oCDT - value to division
	  @return Read/write reference to self
	*/
	CDT & operator/=(const CDT & oCDT);

	// ///////////////////////////////////////////////

	// Operator == ///////////////////////////////////

	/**
	  @brief Comparison operator == for CDT argument
	  @param oValue - value to compare
	  @return true if values are equal
	*/
	bool operator==(const INT_64 & oValue) const;

	/**
	  @brief Comparison operator == for CDT argument
	  @param oValue - value to compare
	  @return true if values are equal
	*/
	bool operator==(const UINT_64 & oValue) const;

	/**
	  @brief Comparison operator == for CDT argument
	  @param oValue - value to compare
	  @return true if values are equal
	*/
	bool operator==(const INT_32 & oValue) const;

	/**
	  @brief Comparison operator == for CDT argument
	  @param oValue - value to compare
	  @return true if values are equal
	*/
	bool operator==(const UINT_32 & oValue) const;

	/**
	  @brief Comparison operator == for CDT argument
	  @param oValue - value to compare
	  @return true if values are equal
	*/
	bool operator==(const W_FLOAT & oValue) const;

	/* *
	  @brief Comparison operator == for string argument (removed, because it's ambigous to bool CTPP::CDT::operator==(const CTPP::CDT&) const)
	  @param oValue - value to compare
	  @return true if values are equal
	*/
	// bool operator==(const std::string & oValue) const;

	/**
	  @brief Comparison operator == for CDT argument
	  @param oCDT - value to compare
	  @return true if values are equal
	*/
	bool operator==(const CDT & oCDT) const;

	// Operator != ///////////////////////////////////

	/**
	  @brief Comparison operator != for CDT argument
	  @param oValue - value to compare
	  @return true if values are not equal
	*/
	bool operator!=(const INT_64 & oValue) const;

	/**
	  @brief Comparison operator != for CDT argument
	  @param oValue - value to compare
	  @return true if values are not equal
	*/
	bool operator!=(const UINT_64 & oValue) const;

	/**
	  @brief Comparison operator != for CDT argument
	  @param oValue - value to compare
	  @return true if values are not equal
	*/
	bool operator!=(const INT_32 & oValue) const;

	/**
	  @brief Comparison operator != for CDT argument
	  @param oValue - value to compare
	  @return true if values are not equal
	*/
	bool operator!=(const UINT_32 & oValue) const;

	/**
	  @brief Comparison operator != for CDT argument
	  @param oValue - value to compare
	  @return true if values are not equal
	*/
	bool operator!=(const W_FLOAT & oValue) const;

	/* *
	  @brief Comparison operator != for string argument (removed, because it's ambigous to bool CTPP::CDT::operator!=(const CTPP::CDT&) const)
	  @param oValue - value to compare
	  @return true if values are not equal
	*/
	// bool operator!=(const std::string & oValue) const;

	/**
	  @brief Comparison operator != for CDT argument
	  @param oCDT - value to compare
	  @return true if values are not equal
	*/
	bool operator!=(const CDT & oCDT) const;
	// ///////////////////////////////////////////////

	// Operator > ///////////////////////////////////

	/**
	  @brief Comparison operator > for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater than value
	*/
	bool operator>(const INT_64 & oValue) const;

	/**
	  @brief Comparison operator > for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater than value
	*/
	bool operator>(const UINT_64 & oValue) const;

	/**
	  @brief Comparison operator > for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater than value
	*/
	bool operator>(const INT_32 & oValue) const;

	/**
	  @brief Comparison operator > for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater than value
	*/
	bool operator>(const UINT_32 & oValue) const;

	/**
	  @brief Comparison operator > for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater than value
	*/
	bool operator>(const W_FLOAT & oValue) const;

	/* *
	  @brief Comparison operator > for string argument (removed, because it's ambigous to bool CTPP::CDT::operator>(const CTPP::CDT&) const)
	  @param oValue - value to compare
	  @return true if CDT is greater than value
	*/
	// bool operator>(const std::string & oValue) const;

	/**
	  @brief Comparison operator > for CDT argument
	  @param oCDT - value to compare
	  @return true if CDT is greater than value
	*/
	bool operator>(const CDT & oCDT) const;

	// Operator < ///////////////////////////////////

	/**
	  @brief Comparison operator < for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is lesser than value
	*/
	bool operator<(const INT_64 & oValue) const;

	/**
	  @brief Comparison operator < for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is lesser than value
	*/
	bool operator<(const UINT_64 & oValue) const;

	/**
	  @brief Comparison operator < for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is lesser than value
	*/
	bool operator<(const INT_32 & oValue) const;

	/**
	  @brief Comparison operator < for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is lesser than value
	*/
	bool operator<(const UINT_32 & oValue) const;

	/**
	  @brief Comparison operator < for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is lesser than value
	*/
	bool operator<(const W_FLOAT & oValue) const;

	/* *
	  @brief Comparison operator < for string argument (removed, because it's ambigous to bool CTPP::CDT::operator<(const CTPP::CDT&) const)
	  @param oValue - value to compare
	  @return true if CDT is lesser than value
	*/
	// bool operator<(const std::string & oValue) const;

	/**
	  @brief Comparison operator < for CDT argument
	  @param oCDT - value to compare
	  @return true if CDT is lesser than value
	*/
	bool operator<(const CDT & oCDT) const;

	// Operator >= ///////////////////////////////////

	/**
	  @brief Comparison operator >= for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater or equal than value
	*/
	bool operator>=(const INT_64 & oValue) const;

	/**
	  @brief Comparison operator >= for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater or equal than value
	*/
	bool operator>=(const UINT_64 & oValue) const;

	/**
	  @brief Comparison operator >= for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater or equal than value
	*/
	bool operator>=(const INT_32 & oValue) const;

	/**
	  @brief Comparison operator >= for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater or equal than value
	*/
	bool operator>=(const UINT_32 & oValue) const;

	/**
	  @brief Comparison operator >= for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater or equal than value
	*/
	bool operator>=(const W_FLOAT & oValue) const;

	/* *
	  @brief Comparison operator >= for string argument (removed, because it's ambigous to bool CTPP::CDT::operator>=(const CTPP::CDT&) const)
	  @param oValue - value to compare
	  @return true if CDT is greater or equal than value
	*/
	// bool operator>=(const std::string & oValue) const;

	/**
	  @brief Comparison operator >= for CDT argument
	  @param oCDT - value to compare
	  @return true if CDT is greater or equal than value
	*/
	bool operator>=(const CDT & oCDT) const;

	// Operator <= ///////////////////////////////////

	/**
	  @brief Comparison operator <= for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater or equal than value
	*/
	bool operator<=(const INT_64 & oValue) const;

	/**
	  @brief Comparison operator <= for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater or equal than value
	*/
	bool operator<=(const UINT_64 & oValue) const;

	/**
	  @brief Comparison operator <= for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater or equal than value
	*/
	bool operator<=(const INT_32 & oValue) const;

	/**
	  @brief Comparison operator <= for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater or equal than value
	*/
	bool operator<=(const UINT_32 & oValue) const;

	/**
	  @brief Comparison operator <= for CDT argument
	  @param oValue - value to compare
	  @return true if CDT is greater or equal than value
	*/
	bool operator<=(const W_FLOAT & oValue) const;

	/* *
	  @brief Comparison operator <= for string argument (removed, because it's ambigous to bool CTPP::CDT::operator<=(const CTPP::CDT&) const)
	  @param oValue - value to compare
	  @return true if CDT is greater or equal than value
	*/
	// bool operator<=(const std::string & oValue) const;

	/**
	  @brief Comparison operator <= for CDT argument
	  @param oCDT - value to compare
	  @return true if CDT is greater or equal than value
	*/
	bool operator<=(const CDT & oCDT) const;

	// ///////////////////////////////////////////////

	/**
	  @brief Pre-increment operator ++
	  @return Read/write reference to self
	*/
	inline CDT & operator++();

	/**
	  @brief Post-increment operator ++
	  @return Read/write reference to self
	*/
	inline CDT   operator++(int);

	/**
	  @brief Pre-decrement operator --
	  @return Read/write reference to self
	*/
	inline CDT & operator--();

	/**
	  @brief Post-decrement operator --
	  @return Read/write reference to self
	*/
	inline CDT   operator--(int);

	/**
	  @brief Append a string to CDT
	  @see Append(const std::string & oValue)
	  @param oValue - string to append
	  @return read/write referense to self
	*/
	inline CDT & Concat(const std::string & oValue);

	/**
	  @brief Append a string to CDT, alias for Concat
	  @see Concat(const std::string & oValue)
	  @param oValue - string to append
	  @return read/write referense to self
	*/
	inline CDT & Append(const std::string & oValue) { return Concat(oValue); }

	/**
	  @brief Append a old-fashion string to CDT
	  @see Append(CCHAR_P * szData, const INT_32 & iDataLength)
	  @param szData - string to append
	  @param iDataLength - string length
	  @return read/write referense to self
	*/
	CDT & Concat(CCHAR_P szData, const INT_32 & iDataLength);

	/**
	  @brief Append a old-fashion string to CDT
	  @see Concat(CCHAR_P * szData, const INT_32 & iDataLength)
	  @param szData - string to append
	  @param iDataLength - string length
	  @return read/write referense to self
	*/
	inline CDT & Append(CCHAR_P szData, const INT_32 & iDataLength) { return Concat(szData, iDataLength); }

	/**
	  @brief Get value as W_FLOAT
	*/
	W_FLOAT GetFloat() const;

	/**
	  @brief Get value as INT_64
	*/
	INT_64 GetInt() const;

	/**
	  @brief Get value as UINT_64
	*/
	UINT_64 GetUInt() const;

	/**
	  @brief Get value as std::string
	  @param szFormat - output format
	  @return String object representation
	*/
	std::string GetString(CCHAR_P szFormat = "") const;

	/**
	  @brief Get generic pointer
	  @return generic pointer
	*/
	void * GetPointer() const;

	/**
	  @brief Get object
	  @return pointer to specified type
	*/
	template<typename T> T * GetObject() const { return (T*)GetPointer(); }

	/**
	  @brief Recursively dump CDT into string
	  @param iLevel - left margin
	  @return printable string
	*/
	std::string RecursiveDump(UINT_32 iLevel = 0) const;

	/**
	  @brief Get value type of object
	*/
	inline eValType GetType() const { return eValueType; }

	/**
	  @brief Get array of hash size
	  @return Size of array
	*/
	inline UINT_32 Size() const
	{
		if      (eValueType == ARRAY_VAL)  { return u.p_data -> u.v_data -> size(); }
		else if (eValueType == HASH_VAL)   { return u.p_data -> u.m_data -> size(); }
		else if (eValueType == STRING_VAL) { return u.p_data -> u.s_data -> size(); }

	return 0;
	}

	/**
	  @brief Get printable value type of object
	  @return printable name of data type
	*/
	inline CCHAR_P PrintableType() const throw() { return PrintableType(eValueType); }

	/**
	  @brief Get printable value type
	  @return printable name of data type
	*/
	inline CCHAR_P PrintableType(eValType eType) const throw()
	{
		switch(eType)
		{
			case UNDEF:       return "*UNDEF*";
			case INT_VAL:     return "INTEGER";
			case REAL_VAL:    return "REAL";
			case STRING_VAL:  return "STRING";
			case ARRAY_VAL:   return "ARRAY";
			case HASH_VAL:    return "HASH";
			case POINTER_VAL: return "POINTER";
			default:          return "!?OMG_WTF?!";
		}
	// This should *NOT* happened;
	return NULL;
	}

	// FWD
	class ConstIterator;

	/**
	  @class Iterator CDT.hpp <CDT.hpp>
	  @brief CDT[HASH] forward iterator
	*/
	class Iterator
	{
	private:
		friend class CDT;
		friend class ConstIterator;

		/** Hash iterator */
		CDT::Map::iterator itMap;

		/**
		  @brief Constructor
		  @param itIMap - map iterator
		*/
		inline Iterator(CDT::Map::iterator itIMap): itMap(itIMap) { ;; }
	public:
		/**
		  @brief Copy constructor
		  @param oRhs - object to copy
		*/
		inline Iterator(const Iterator & oRhs): itMap(oRhs.itMap) { ;; }

		/**
		  @brief Operator =
		  @param oRhs - object to copy
		*/
		inline Iterator & operator=(const Iterator & oRhs)
		{
			if (this != &oRhs) { itMap = oRhs.itMap; }

		return *this;
		}

		/**
		  @brief Pre-increment operator ++
		*/
		inline Iterator & operator++()
		{
			++itMap;

		return *this;
		}

		/**
		  @brief Post-increment operator ++
		*/
		inline Iterator operator++(int)
		{
			Iterator oTMP = *this;

			++itMap;

		return oTMP;
		}

		/**
		  @brief Access operator
		  @return Pair of key => value
		*/
		inline std::pair<const std::string, CDT> * operator->() { return &(*itMap); }

		/**
		  @brief Comparison operator
		  @param oRhs - object to compare
		  @return true if objects are equal
		*/
		inline bool operator ==(const Iterator & oRhs) { return (itMap == oRhs.itMap); }

		/**
		  @brief Comparison operator
		  @param oRhs - object to compare
		  @return true if objects are NOT equal
		*/
		inline bool operator !=(const Iterator & oRhs) { return (itMap != oRhs.itMap); }
	};

	/**
	  @brief Get iterator pointed to start of hash
	*/
	inline Iterator Begin()
	{
		if (eValueType != HASH_VAL) { throw CDTAccessException(); }

		return Iterator(u.p_data -> u.m_data -> begin());
	}

	/**
	  @brief Get iterator pointed to end of hash
	*/
	inline Iterator End()
	{
		if (eValueType != HASH_VAL) { throw CDTAccessException(); }

		return Iterator(u.p_data -> u.m_data -> end());
	}

	/**
	  @brief Find element in hash
	  @param sKey - element name
	  @return Iterator pointed to element or to end of hash if nothing found
	*/
	inline Iterator Find(const std::string & sKey)
	{
		if (eValueType != HASH_VAL) { throw CDTAccessException(); }

		return Iterator(u.p_data -> u.m_data -> find(sKey));
	}

	/**
	  @class ConstIterator CDT.hpp <CDT.hpp>
	  @brief CDT[HASH] forward constant iterator
	*/
	class ConstIterator
	{
	private:
		friend class CDT;

		/** Hash iterator */
		CDT::Map::const_iterator itMap;

	public:
		/**
		  @brief Copy constructor
		  @param oRhs - object to copy
		*/
		inline ConstIterator(const ConstIterator & oRhs): itMap(oRhs.itMap) { ;; }

		/**
		  @brief Type cast constructor
		  @param oRhs - object to copy
		*/
		inline ConstIterator(const Iterator & oRhs): itMap(oRhs.itMap) { ;; }

		/**
		  @brief Operator =
		  @param oRhs - object to copy
		*/
		inline ConstIterator & operator=(const ConstIterator & oRhs)
		{
			if (this != &oRhs) { itMap = oRhs.itMap; }

		return *this;
		}

		/**
		  @brief Operator =
		  @param oRhs - object to copy
		*/
		inline ConstIterator & operator=(const Iterator & oRhs)
		{
			itMap = oRhs.itMap;

		return *this;
		}

		/**
		  @brief Pre-increment operator ++
		*/
		inline ConstIterator & operator++()
		{
			++itMap;

		return *this;
		}

		/**
		  @brief Post-increment operator ++
		*/
		inline ConstIterator operator++(int)
		{
			ConstIterator oTMP = *this;

			++itMap;

		return oTMP;
		}

		/**
		  @brief Access operator
		  @return Pair of key => value
		*/
		inline const std::pair<const std::string, CDT> * operator->() const { return &(*itMap); }

		/**
		  @brief Comparison operator
		  @param oRhs - object to compare
		  @return true if objects are equal
		*/
		inline bool operator ==(const ConstIterator & oRhs) const { return (itMap == oRhs.itMap); }

		/**
		  @brief Comparison operator
		  @param oRhs - object to compare
		  @return true if objects are NOT equal
		*/
		inline bool operator !=(const ConstIterator & oRhs) const { return (itMap != oRhs.itMap); }
	};

	/**
	  @brief Get constant iterator pointed to start of hash
	*/
	inline ConstIterator Begin() const
	{
		if (eValueType != HASH_VAL) { throw CDTAccessException(); }

		return ConstIterator(u.p_data -> u.m_data -> begin());
	}

	/**
	  @brief Get constant iterator pointed to end of hash
	*/
	inline ConstIterator End() const
	{
		if (eValueType != HASH_VAL) { throw CDTAccessException(); }

		return ConstIterator(u.p_data -> u.m_data -> end());
	}

	/**
	  @brief Find element in hash
	  @param sKey - element name
	  @return Iterator pointed to element or to end of hash if nothing found
	*/
	inline ConstIterator Find(const std::string & sKey) const
	{
		if (eValueType != HASH_VAL) { throw CDTAccessException(); }

		return ConstIterator(u.p_data -> u.m_data -> find(sKey));
	}

	/**
	  @brief A destructor
	*/
	~CDT() throw();

private:
	// Friends
	friend CDT operator-(const UINT_64 & oValue, const CDT & oCDT);
	friend CDT operator-(const INT_64  & oValue, const CDT & oCDT);
	friend CDT operator-(const W_FLOAT & oValue, const CDT & oCDT);

	friend CDT operator/(const UINT_64 & oValue, const CDT & oCDT);
	friend CDT operator/(const INT_64  & oValue, const CDT & oCDT);
	friend CDT operator/(const W_FLOAT & oValue, const CDT & oCDT);

	/**
	  @struct _CDT CDT.hpp <CDT.hpp>
	  @brief Shareable datatypes
	*/
	struct _CDT
	{
		/** Union */
		union
		{
			/** String */
			std::string      * s_data;
			/** Array  */
			Vector           * v_data;
			/** Hash   */
			Map              * m_data;
		} u;

		/** References counter */
		UINT_32 refcount;

		/** Constructor */
		_CDT(): refcount(1) { u.s_data = NULL; }
	};

	/** Plain Old datatypes */
	union
	{
		/** Signed interger                */
		INT_64                 i_data;
		/** Floating point value           */
		W_FLOAT                d_data;
		/** Pointer to shareable container */
		_CDT                 * p_data;
		/** Generic pointer                */
		void                 * pp_data;
	} u;

	/** Value type */
	eValType              eValueType;

	/**
	  @brief Destroy object if need
	*/
	void Destroy() throw();

	/**
	  @brief Unshare shareable container
	*/
	void Unshare();

	/**
	  @brief Dump CDT into string
	  @param iLevel  - level of recursion
	  @param oData   - data to dump
	  @param sResult - string to put result in
	*/
	void DumpData(UINT_32 iLevel, const CDT & oData, std::string & sResult) const;
};

// ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Realization
//

//
// Pre-increment operator ++
//
inline CDT & CDT::operator++()
{
	if (eValueType == INT_VAL) { ++u.i_data; }
	else { throw CDTTypeCastException("pre-increment operator ++"); }
return *this;
}

//
// Post-increment operator ++
//
inline CDT CDT::operator++(int)
{
	CDT oTMP(*this);

	if (eValueType == INT_VAL) { ++u.i_data; }
	else { throw CDTTypeCastException("post-increment operator ++"); }

return oTMP;
}

//
// Pre-decrement operator --
//
inline CDT & CDT::operator--()
{
	if (eValueType == INT_VAL) { --u.i_data; }
	else { throw CDTTypeCastException("pre-decrement operator --"); }

return *this;
}

//
// Post-decrement operator --
//
inline CDT CDT::operator--(int)
{
	CDT oTMP(*this);

	if (eValueType == INT_VAL) { --u.i_data; }
	else { throw CDTTypeCastException("post-decrement operator --"); }

return oTMP;
}

// ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Concatenate two values
//
inline CDT & CDT::Concat(const std::string & oValue) { return this -> operator+=(oValue); }

// ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
  @brief Operator + for UINT_64 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of addition with same type of oCDT
*/
inline CDT operator+(const UINT_64 & oValue, const CDT & oCDT) { return CDT(oValue).operator+(oCDT); }

/**
  @brief Operator + for INT_64 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of addition with same type of oCDT
*/
inline CDT operator+(const INT_64 & oValue, const CDT & oCDT)  { return CDT(oValue).operator+(oCDT); }

/**
  @brief Operator + for UINT_32 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of addition with same type of oCDT
*/
inline CDT operator+(const UINT_32 & oValue, const CDT & oCDT) { return CDT(oValue).operator+(oCDT); }

/**
  @brief Operator + for INT_32 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of addition with same type of oCDT
*/
inline CDT operator+(const INT_32 & oValue, const CDT & oCDT) { return CDT(oValue).operator+(oCDT); }

/**
  @brief Operator + for W_FLOAT and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of addition with same type of oCDT
*/
inline CDT operator+(const W_FLOAT & oValue, const CDT & oCDT) { return CDT(oValue).operator+(oCDT); }

// ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
  @brief Operator - for UINT_64 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of substraction with same type of oCDT
*/
inline CDT operator-(const UINT_64 & oValue, const CDT & oCDT)
{
	if      (oCDT.eValueType == CDT::INT_VAL)  { return CDT(oValue - oCDT.u.i_data); }
	else if (oCDT.eValueType == CDT::REAL_VAL) { return CDT(oValue - oCDT.u.d_data); }
	else                                       { throw CDTTypeCastException("standalone operator -"); }

return CDT(oValue);
}

/**
  @brief Operator - for INT_64 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of substraction with same type of oCDT
*/
inline CDT operator-(const INT_64 & oValue, const CDT & oCDT)
{
	if      (oCDT.eValueType == CDT::INT_VAL)  { return CDT(oValue - oCDT.u.i_data); }
	else if (oCDT.eValueType == CDT::REAL_VAL) { return CDT(oValue - oCDT.u.d_data); }
	else                                       { throw CDTTypeCastException("standalone operator -"); }

return CDT(oValue);
}

/**
  @brief Operator - for UINT_32 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of substraction with same type of oCDT
*/
inline CDT operator-(const UINT_32 & oValue, const CDT & oCDT) { return operator-(UINT_64(oValue), oCDT); }

/**
  @brief Operator - for INT_32 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of substraction with same type of oCDT
*/
inline CDT operator-(const INT_32 & oValue, const CDT & oCDT)  { return operator-(INT_64(oValue), oCDT); }

/**
  @brief Operator - for W_FLOAT and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of substraction with same type of oCDT
*/
inline CDT operator-(const W_FLOAT & oValue, const CDT & oCDT)
{
	if      (oCDT.eValueType == CDT::INT_VAL)  { return CDT(oValue - oCDT.u.i_data); }
	else if (oCDT.eValueType == CDT::REAL_VAL) { return CDT(oValue - oCDT.u.d_data); }
	else                                       { throw CDTTypeCastException("standalone operator -"); }

return CDT(oValue);
}

// ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
  @brief Operator * for UINT_64 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of multiplication with same type of oCDT
*/
inline CDT operator*(const UINT_64 & oValue, const CDT & oCDT) { return CDT(oCDT).operator*(oValue); }

/**
  @brief Operator * for INT_64 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of multiplication with same type of oCDT
*/
inline CDT operator*(const INT_64 & oValue, const CDT & oCDT)  { return CDT(oCDT).operator*(oValue); }

/**
  @brief Operator * for UINT_32 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of multiplication with same type of oCDT
*/
inline CDT operator*(const UINT_32 & oValue, const CDT & oCDT) { return CDT(oCDT).operator*(oValue); }

/**
  @brief Operator * for INT_32 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of multiplication with same type of oCDT
*/
inline CDT operator*(const INT_32 & oValue, const CDT & oCDT)  { return CDT(oCDT).operator*(oValue); }

/**
  @brief Operator * for W_FLOAT and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of multiplication with same type of oCDT
*/
inline CDT operator*(const W_FLOAT & oValue, const CDT & oCDT) { return CDT(oCDT).operator*(oValue); }

// ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
  @brief Operator / for UINT_64 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of division with same type of oCDT
*/
inline CDT operator/(const UINT_64 & oValue, const CDT & oCDT)
{
	if      (oCDT.eValueType == CDT::INT_VAL)  { return CDT(oValue / oCDT.u.i_data); }
	else if (oCDT.eValueType == CDT::REAL_VAL) { return CDT(oValue / oCDT.u.d_data); }
	else                                       { throw CDTTypeCastException("standalone operator /"); }

return CDT(oValue);
}

/**
  @brief Operator / for INT_64 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of division with same type of oCDT
*/
inline CDT operator/(const INT_64 & oValue, const CDT & oCDT)
{
	if      (oCDT.eValueType == CDT::INT_VAL)  { return CDT(oValue / oCDT.u.i_data); }
	else if (oCDT.eValueType == CDT::REAL_VAL) { return CDT(oValue / oCDT.u.d_data); }
	else                                       { throw CDTTypeCastException("standalone operator /"); }

return CDT(oValue);
}

/**
  @brief Operator / for UINT_32 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of division with same type of oCDT
*/
inline CDT operator/(const UINT_32 & oValue, const CDT & oCDT) { return operator/(UINT_64(oValue), oCDT); }

/**
  @brief Operator / for INT_32 and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of division with same type of oCDT
*/
inline CDT operator/(const INT_32 & oValue, const CDT & oCDT)  { return operator/(INT_64(oValue), oCDT); }

/**
  @brief Operator / for W_FLOAT and CDT arguments
  @param oValue - 1-st argument
  @param oCDT - 2-nd argument
  @return result of division with same type of oCDT
*/
inline CDT operator/(const W_FLOAT & oValue, const CDT & oCDT)
{
	if      (oCDT.eValueType == CDT::INT_VAL)  { return CDT(oValue / oCDT.u.i_data); }
	else if (oCDT.eValueType == CDT::REAL_VAL) { return CDT(oValue / oCDT.u.d_data); }
	else                                       { throw CDTTypeCastException("standalone operator /"); }

return CDT(oValue);
}

// ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

} // namespace CTPP
#endif // _CDT_HPP__
// End.
