/*-
 * Copyright (c) 2004 - 2010 CTPP Team
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the CTPP Team nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      CDT.hpp
 *
 * $CTPP$
 */
#include "CDT.hpp"

#include <stdio.h>

namespace CTPP
{
// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Default constructor
//
CDT::CDT(const eValType & oValue): eValueType(oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			break;

		case INT_VAL:
			u.i_data = 0;
			break;

		case REAL_VAL:
			u.d_data = 0;
			break;

		case STRING_VAL:
			u.p_data = new _CDT();
			u.p_data -> u.s_data = new std::string;
			break;

		case ARRAY_VAL:
			u.p_data = new _CDT();
			u.p_data -> u.v_data = new Vector;
			break;

		case HASH_VAL:
			u.p_data = new _CDT();
			u.p_data -> u.m_data = new Map;
			break;

		case POINTER_VAL:
			u.pp_data = NULL;
			break;

		default:
			throw CDTTypeCastException("No such type");
	}
}

//
// Copy constructor
//
CDT::CDT(const CDT & oCDT)
{
	if (this == &oCDT) { eValueType = UNDEF; }
	else
	{
		eValueType = oCDT.eValueType;
		switch (eValueType)
		{
			case UNDEF:
				break;

			case INT_VAL:
				u.i_data = oCDT.u.i_data;
				break;

			case REAL_VAL:
				u.d_data = oCDT.u.d_data;
				break;

			case STRING_VAL:
			case ARRAY_VAL:
			case HASH_VAL:
				u.p_data = oCDT.u.p_data;
				++(u.p_data -> refcount);
				break;

			default:
				throw CDTTypeCastException("No such type");
		}
	}
}

//
// Copy operator
//
CDT & CDT::operator=(const CDT & oCDT)
{
	if (this != &oCDT)
	{
		// Destroy object if need
		if (eValueType >= STRING_VAL) { Destroy(); }

		eValueType = oCDT.eValueType;

		switch (eValueType)
		{
			case UNDEF:
				break;

			case INT_VAL:
				u.i_data = oCDT.u.i_data;
				break;

			case REAL_VAL:
				u.d_data = oCDT.u.d_data;
				break;

			case STRING_VAL:
			case ARRAY_VAL:
			case HASH_VAL:
				u.p_data = oCDT.u.p_data;
				++u.p_data -> refcount;
				break;

			case POINTER_VAL:
				u.pp_data = oCDT.u.pp_data;
				break;

			default:
				throw CDTTypeCastException("No such type");
		}
	}

return *this;
}

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Copy constructor from INT_64 type
//
CDT::CDT(const INT_64 & oValue): eValueType(INT_VAL)        { u.i_data = oValue; }

//
// Copy constructor from UINT_64 type
//
CDT::CDT(const UINT_64 & oValue): eValueType(INT_VAL)       { u.i_data = oValue; }

//
// Copy constructor from INT_32 type
//
CDT::CDT(const INT_32 & oValue): eValueType(INT_VAL)        { u.i_data = oValue; }

//
// Copy constructor from UINT_32 type
//
CDT::CDT(const UINT_32 & oValue): eValueType(INT_VAL)       { u.i_data = oValue; }

//
// Copy constructor from W_FLOAT type
//
CDT::CDT(const W_FLOAT & oValue): eValueType(REAL_VAL)      { u.d_data = oValue; }

//
// Copy constructor from std::string type
//
CDT::CDT(const std::string & oValue): eValueType(STRING_VAL)
{
	u.p_data = new _CDT();
	u.p_data -> u.s_data = new std::string(oValue);
}

//
// Copy constructor from CCHAR_P type
//
CDT::CDT(CCHAR_P oValue): eValueType(STRING_VAL)
{
	u.p_data = new _CDT();
	u.p_data -> u.s_data = new std::string(oValue);
}

//
// Type cast constructor
//
CDT::CDT(void * oValue): eValueType(POINTER_VAL) { u.pp_data = oValue; }

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator = for type INT_64
//
CDT & CDT::operator=(const INT_64 & oValue)
{
	// Destroy object if need
	if (eValueType >= STRING_VAL) { Destroy(); }

	eValueType = INT_VAL;
	u.i_data   = oValue;

return *this;
}

//
// Operator = for type UINT_64
//
CDT & CDT::operator=(const UINT_64 & oValue)
{
	// Destroy object if need
	if (eValueType >= STRING_VAL) { Destroy(); }

	eValueType = INT_VAL;
	u.i_data   = oValue;

return *this;
}

//
// Operator = for type INT_32
//
CDT & CDT::operator=(const INT_32 & oValue)
{
	// Destroy object if need
	if (eValueType >= STRING_VAL) { Destroy(); }

	eValueType = INT_VAL;
	u.i_data   = oValue;

return *this;
}

//
// Operator = for type UINT_32
//
CDT & CDT::operator=(const UINT_32 & oValue)
{
	// Destroy object if need
	if (eValueType >= STRING_VAL) { Destroy(); }

	eValueType = INT_VAL;
	u.i_data   = oValue;

return *this;
}

//
// Operator = for type W_FLOAT
//
CDT & CDT::operator=(const W_FLOAT & oValue)
{
	// Destroy object if need
	if (eValueType >= STRING_VAL) { Destroy(); }

	eValueType = REAL_VAL;
	u.d_data   = oValue;

return *this;
}

//
// Operator = for type std::string
//
CDT & CDT::operator=(const std::string & oValue)
{
	// Destroy object if need
	if (eValueType >= STRING_VAL) { Destroy(); }

	eValueType = STRING_VAL;
	u.p_data = new _CDT();
	u.p_data -> u.s_data = new std::string(oValue);

return *this;
}

//
// Operator = for type std::string
//
CDT & CDT::operator=(CCHAR_P oValue)
{
	// Destroy object if need
	if (eValueType >= STRING_VAL) { Destroy(); }

	eValueType = STRING_VAL;
	u.p_data = new _CDT();
	u.p_data -> u.s_data = new std::string(oValue);

return *this;
}

//
// Type cast constructor
//
CDT & CDT::operator=(void * oValue)
{
	// Destroy object if need
	if (eValueType >= STRING_VAL) { Destroy(); }

	eValueType = POINTER_VAL; 
	u.pp_data  = oValue;

return *this;
}

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Access operator []
//
CDT & CDT::operator[](const UINT_32 & iPos)
{
	if (eValueType == UNDEF)
	{
		eValueType = ARRAY_VAL;
		u.p_data = new _CDT;
		u.p_data -> u.v_data = new Vector(iPos + 1);
	}
	else if (eValueType != ARRAY_VAL) { throw CDTAccessException(); }

	// Unshare complex type
	Unshare();

	UINT_32 iSize = u.p_data -> u.v_data -> size();

	if      (iPos == iSize) { u.p_data -> u.v_data -> push_back(CDT()); }
	else if (iPos > iSize)  { u.p_data -> u.v_data -> resize(iPos + 1); }

return u.p_data -> u.v_data -> operator[](iPos);
}

//
// Access operator []
//
CDT & CDT::operator[](const std::string & sKey)
{
	if (eValueType == UNDEF)
	{
		eValueType = HASH_VAL;
		u.p_data = new _CDT;
		u.p_data -> u.m_data = new Map;
	}
	else if (eValueType != HASH_VAL) { throw CDTAccessException(); }

	// Unshare complex type
	Unshare();

return u.p_data -> u.m_data -> operator[](sKey);
}

//
// Provides constant access to the data contained in CDT
//
CDT CDT::GetCDT(const UINT_32 & iPos) const
{
	if (eValueType != ARRAY_VAL) { throw CDTAccessException(); }

	if (iPos >= u.p_data -> u.v_data -> size()) { throw CDTRangeException(); }

return CDT(u.p_data -> u.v_data -> operator[](iPos));
}

//
// Provides constant access to the data contained in CDT
//
CDT CDT::GetCDT(const std::string & sKey) const
{
	INT_32 iFlag = 0;

return GetExistedCDT(sKey, iFlag);
}

//
// Provides constant access to the data contained in CDT
//
CDT CDT::GetExistedCDT(const std::string & sKey, INT_32 & iCDTExist) const
{
	// CDT Does Not exist
	if (eValueType != HASH_VAL) { iCDTExist = -1; return CDT(); }

	Map::const_iterator itmHash = u.p_data -> u.m_data -> find(sKey);
	if (itmHash == u.p_data -> u.m_data -> end())
	{
		iCDTExist = -1;
		return CDT();
	}
	iCDTExist = 0;

return CDT(itmHash -> second);
}

//
// Erase element from HASH
//
bool CDT::Erase(const std::string & sKey)
{
	if (eValueType != HASH_VAL) { throw CDTAccessException(); }

	Map::iterator itmHash = u.p_data -> u.m_data -> find(sKey);
	if (itmHash == u.p_data -> u.m_data -> end()) { return false; }

	Unshare();

	u.p_data -> u.m_data -> erase(itmHash);

return true;
}

//
// Check element in array
//
bool CDT::Exists(const UINT_32 & iPos) const
{
	if (eValueType != ARRAY_VAL) { throw CDTAccessException(); }

	if (iPos >= u.p_data -> u.v_data -> size()) { return false; }

return true;	
}

//
// Check element in hash
//
bool CDT::Exists(const std::string & sKey) const
{
	if (eValueType != HASH_VAL) { throw CDTAccessException(); }

	Map::const_iterator itmHash = u.p_data -> u.m_data -> find(sKey);
	if (itmHash == u.p_data -> u.m_data -> end()) { return false; }

return true;
}
//
// Push value into array
//
void CDT::PushBack(const INT_64 & oValue)
{
	if (eValueType != ARRAY_VAL) { throw CDTAccessException(); }

	u.p_data -> u.v_data -> push_back(oValue);
}

//
// Push value into array
//
void CDT::PushBack(const UINT_64 & oValue)
{
	if (eValueType != ARRAY_VAL) { throw CDTAccessException(); }

	u.p_data -> u.v_data -> push_back(oValue);
}

//
// Push value into array
//
void CDT::PushBack(const INT_32 & oValue)
{
	if (eValueType != ARRAY_VAL) { throw CDTAccessException(); }

	u.p_data -> u.v_data -> push_back(oValue);
}

//
// Push value into array
//
void CDT::PushBack(const UINT_32 & oValue)
{
	if (eValueType != ARRAY_VAL) { throw CDTAccessException(); }

	u.p_data -> u.v_data -> push_back(oValue);
}

//
// Push value into array
//
void CDT::PushBack(const W_FLOAT & oValue)
{
	if (eValueType != ARRAY_VAL) { throw CDTAccessException(); }

	u.p_data -> u.v_data -> push_back(oValue);
}

//
// Push value into array
//
void CDT::PushBack(const std::string & oValue)
{
	if (eValueType != ARRAY_VAL) { throw CDTAccessException(); }

	u.p_data -> u.v_data -> push_back(oValue);
}

//
// Push value into array
//
void CDT::PushBack(CCHAR_P oValue)
{
	if (eValueType != ARRAY_VAL) { throw CDTAccessException(); }

	u.p_data -> u.v_data -> push_back(oValue);
}

//
// Push value into array
//
void CDT::PushBack(const CDT & oValue)
{
	if (eValueType != ARRAY_VAL) { throw CDTAccessException(); }

	u.p_data -> u.v_data -> push_back(oValue);
}

//
// Returns a boolean value telling whether object has a value
//
bool CDT::Defined()
{
	switch (eValueType)
	{
		case UNDEF:
			break;

		case INT_VAL:
			if (u.i_data != 0) { return true; }
			break;

		case REAL_VAL:
			if (u.d_data != 0) { return true; }
			break;

		case STRING_VAL:
			if (u.p_data -> u.s_data -> size() != 0) { return true; }
			break;

		case ARRAY_VAL:
			if (u.p_data -> u.v_data -> size() != 0) { return true; }
			break;

		case HASH_VAL:
			if (u.p_data -> u.m_data -> size() != 0) { return true; }
			break;

		case POINTER_VAL:
			if (u.pp_data != NULL)                   { return true; }
			break;
	}

return false;
}

//
// Provides range-check access to the data contained in CDT
//
CDT & CDT::At(const UINT_32 & iPos)
{
	if (eValueType != ARRAY_VAL) { throw CDTAccessException(); }

	if (iPos >= u.p_data -> u.v_data -> size()) { throw CDTRangeException(); }

return u.p_data -> u.v_data -> operator[](iPos);
}

//
// Provides range-check access to the data contained in CDT
//
CDT & CDT::At(const std::string & sKey)
{
	if (eValueType != HASH_VAL) { throw CDTAccessException(); }

	Map::iterator itmHash = u.p_data -> u.m_data -> find(sKey);
	if (itmHash == u.p_data -> u.m_data -> end()) { throw CDTRangeException(); }

return itmHash -> second;
}

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator + for type INT_64
//
CDT CDT::operator+(const INT_64 & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			return CDT(oValue);

		case INT_VAL:
			return CDT(u.i_data + oValue);

		case REAL_VAL:
			return CDT(u.d_data + oValue);

		case STRING_VAL:
			{
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];
				snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "%lli", (long long) (oValue));
				return CDT(*(u.p_data -> u.s_data) + szBuf);
			}

		default:
			throw CDTTypeCastException("operator + (INT_64)");
	}
// Make stupid compiler happy
return CDT(*this);
}

//
// Operator + for type UINT_64
//
CDT CDT::operator+(const UINT_64 & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			return CDT(oValue);

		case INT_VAL:
			return CDT(u.i_data + oValue);

		case REAL_VAL:
			return CDT(u.d_data + oValue);

		case STRING_VAL:
			{
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];
				snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "%llu", (long long unsigned) (oValue));
				return CDT(*(u.p_data -> u.s_data) + szBuf);
			}
		default:
			throw CDTTypeCastException("operator + (UINT_64)");
	}
// Make stupid compiler happy
return CDT(*this);
}

//
// Operator + for type INT_32
//
CDT CDT::operator+(const INT_32 & oValue)  { return this -> operator+(INT_64(oValue));  }

//
// Operator + for type UINT_32
//
CDT CDT::operator+(const UINT_32 & oValue) { return this -> operator+(UINT_64(oValue));  }

//
// Operator + for type W_FLOAT
//
CDT CDT::operator+(const W_FLOAT & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			return CDT(oValue);

		case INT_VAL:
			return CDT(u.i_data + oValue);

		case REAL_VAL:
			return CDT(u.d_data + oValue);

		case STRING_VAL:
			{
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];
				snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "%.*G", CTPP_FLOAT_PRECISION, oValue);
				return CDT(*(u.p_data -> u.s_data) + szBuf);
			}
		default:
			throw CDTTypeCastException("operator + (W_FLOAT)");
	}
// Make stupid compiler happy
return CDT(*this);
}

//
// Operator + for type CDT
//
CDT CDT::operator+(const CDT & oCDT)
{
	switch (eValueType)
	{
		case UNDEF:
			if (oCDT.eValueType <= STRING_VAL) { return CDT(oCDT); }
			else { throw CDTTypeCastException("operator + (oCDT)"); }

		case INT_VAL:
			if      (oCDT.eValueType == UNDEF)    { return CDT(u.i_data);                 }
			else if (oCDT.eValueType == INT_VAL)  { return CDT(u.i_data + oCDT.u.i_data); }
			else if (oCDT.eValueType == REAL_VAL) { return CDT(u.i_data + oCDT.u.d_data); }
			else if (oCDT.eValueType == STRING_VAL)
			{
				return CDT(this -> GetString() +  *(oCDT.u.p_data -> u.s_data));
			}
			else { throw CDTTypeCastException("operator + (oCDT)"); }

		case REAL_VAL:
			if      (oCDT.eValueType == UNDEF)    { return CDT(u.d_data);                 }
			else if (oCDT.eValueType == INT_VAL)  { return CDT(u.d_data + oCDT.u.i_data); }
			else if (oCDT.eValueType == REAL_VAL) { return CDT(u.d_data + oCDT.u.d_data); }
			else if (oCDT.eValueType == STRING_VAL)
			{
				return CDT(this -> GetString() +  *(oCDT.u.p_data -> u.s_data));
			}
			else { throw CDTTypeCastException("operator + (oCDT)"); }

		case STRING_VAL:
			if      (oCDT.eValueType == UNDEF)    { return CDT(*(u.p_data -> u.s_data));                    }
			else if (oCDT.eValueType == INT_VAL)  { return CDT(*(u.p_data -> u.s_data) + oCDT.GetString()); }
			else if (oCDT.eValueType == REAL_VAL) { return CDT(*(u.p_data -> u.s_data) + oCDT.GetString()); }
			else if (oCDT.eValueType == STRING_VAL)
			{
				return CDT( *(u.p_data -> u.s_data) + *(oCDT.u.p_data -> u.s_data) );
			}
			else { throw CDTTypeCastException("operator + (oCDT)"); }

		default:
			throw CDTTypeCastException("operator + (CDT)");
	}
// Make stupid compiler happy
return CDT(*this);
}
#if (0)
//
// Operator + for type std::string
//
CDT CDT::operator+(const std::string & oValue)
{
	if      (eValueType == UNDEF)      { return CDT(oValue);                           }
	else if (eValueType == INT_VAL ||
	         eValueType == REAL_VAL)   { return CDT(GetString() + oValue);             }
	else if (eValueType == STRING_VAL) { return CDT(*(u.p_data -> u.s_data) + oValue); }
	else                               { throw CDTTypeCastException("operator + (string)");     }

return CDT(*this);
}
#endif
// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator - for type INT_64
//
CDT CDT::operator-(const INT_64 & oValue)  { return this -> operator+(-oValue);  }

//
// Operator - for type UINT_64
//
CDT CDT::operator-(const UINT_64 & oValue) { return this -> operator+(- INT_64(oValue));  }

//
// Operator - for type INT_32
//
CDT CDT::operator-(const INT_32 & oValue)  { return this -> operator+(-oValue);  }

//
// Operator - for type UINT_32
//
CDT CDT::operator-(const UINT_32 & oValue) { return this -> operator+(- INT_32(oValue));  }

//
// Operator - for type W_FLOAT
//
CDT CDT::operator-(const W_FLOAT & oValue) { return this -> operator+(-oValue);  }

//
// Operator - for type CDT
//
CDT CDT::operator-(const CDT & oCDT)
{
	switch (eValueType)
	{
		case UNDEF:
			if      (oCDT.eValueType == UNDEF)    { return CDT(UNDEF);          }
			else if (oCDT.eValueType == INT_VAL)  { return CDT(-oCDT.u.i_data); }
			else if (oCDT.eValueType == REAL_VAL) { return CDT(-oCDT.u.d_data); }
			else                                  { throw CDTTypeCastException("operator - (oCDT)");     }

		case INT_VAL:
			if      (oCDT.eValueType == UNDEF)    { return CDT(u.i_data);                 }
			else if (oCDT.eValueType == INT_VAL)  { return CDT(u.i_data - oCDT.u.i_data); }
			else if (oCDT.eValueType == REAL_VAL) { return CDT(u.i_data - oCDT.u.d_data); }
			else                                  { throw CDTTypeCastException("operator - (oCDT)");     }

		case REAL_VAL:
			if      (oCDT.eValueType == UNDEF)    { return CDT(u.d_data);                 }
			else if (oCDT.eValueType == INT_VAL)  { return CDT(u.d_data - oCDT.u.i_data); }
			else if (oCDT.eValueType == REAL_VAL) { return CDT(u.d_data - oCDT.u.d_data); }
			else                                  { throw CDTTypeCastException("operator - (oCDT)");     }
		default:
			throw CDTTypeCastException("operator - (CDT)");
	}

// Make stupid compiler happy
return CDT(*this);
}

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator * for type INT_64
//
CDT CDT::operator*(const INT_64 & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			return CDT(INT_64(0));

		case INT_VAL:
			return CDT(u.i_data * oValue);

		case REAL_VAL:
			return CDT(u.d_data * oValue);

		default:
			throw CDTTypeCastException("operator * (INT_64)");
	}

// Make stupid compiler happy
return CDT(*this);
}

//
// Operator * for type UINT_64
//
CDT CDT::operator*(const UINT_64 & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			return CDT(UINT_64(0));

		case INT_VAL:
			return CDT(u.i_data * oValue);

		case REAL_VAL:
			return CDT(u.d_data * oValue);

		default:
			throw CDTTypeCastException("operator * (UINT_64)");
	}

// Make stupid compiler happy
return CDT(*this);
}

//
// Operator * for type INT_32
//
CDT CDT::operator*(const INT_32 & oValue)  { return this -> operator*(INT_64(oValue));  }

//
// Operator * for type UINT_32
//
CDT CDT::operator*(const UINT_32 & oValue) { return this -> operator*(UINT_64(oValue));  }

//
// Operator * for type W_FLOAT
//
CDT CDT::operator*(const W_FLOAT & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			return CDT(W_FLOAT(0.0));

		case INT_VAL:
			return CDT(u.i_data * oValue);

		case REAL_VAL:
			return CDT(u.d_data * oValue);

		default:
			throw CDTTypeCastException("operator * (W_FLOAT)");
	}

// Make stupid compiler happy
return CDT(*this);
}

//
// Operator * for type CDT
//
CDT CDT::operator*(const CDT & oCDT)
{
	switch (eValueType)
	{
		case UNDEF:
			return CDT(INT_64(0));

		case INT_VAL:
			if      (oCDT.eValueType == UNDEF)    { return CDT(INT_64(0));                    }
			else if (oCDT.eValueType == INT_VAL)  { return CDT(u.i_data * oCDT.u.i_data);     }
			else if (oCDT.eValueType == REAL_VAL) { return CDT(u.i_data * oCDT.u.d_data);     }
			else                                  { throw CDTTypeCastException("operator * (oCDT)"); }

		case REAL_VAL:
			if      (oCDT.eValueType == UNDEF)    { return CDT(W_FLOAT(0.0));                 }
			else if (oCDT.eValueType == INT_VAL)  { return CDT(u.d_data * oCDT.u.i_data);     }
			else if (oCDT.eValueType == REAL_VAL) { return CDT(u.d_data * oCDT.u.d_data);     }
			else                                  { throw CDTTypeCastException("operator * (oCDT)"); }

		default:
			throw CDTTypeCastException("operator * (CDT)");
	}

// Make stupid compiler happy
return CDT(*this);
}

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator / for type INT_64
//
CDT CDT::operator/(const INT_64 & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			return CDT(0 / oValue);

		case INT_VAL:
			return CDT(u.i_data / oValue);

		case REAL_VAL:
			return CDT(u.d_data / oValue);

		default:
			throw CDTTypeCastException("operator /(INT_64)");
	}

// Make stupid compiler happy
return CDT(*this);
}

//
// Operator / for type UINT_64
//
CDT CDT::operator/(const UINT_64 & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			return CDT(0 / oValue);

		case INT_VAL:
			return CDT(u.i_data / oValue);

		case REAL_VAL:
			return CDT(u.d_data / oValue);

		default:
			throw CDTTypeCastException("operator / (UINT_64)");
	}

// Make stupid compiler happy
return CDT(*this);
}

//
// Operator / for type INT_32
//
CDT CDT::operator/(const INT_32 & oValue)  { return this -> operator/(INT_64(oValue));  }

//
// Operator / for type UINT_32
//
CDT CDT::operator/(const UINT_32 & oValue) { return this -> operator/(UINT_64(oValue));  }

//
// Operator / for type W_FLOAT
//
CDT CDT::operator/(const W_FLOAT & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			// Keep this
			// For explanation see IEEE standard 754 for floating-point arithmetic
			return CDT(0.0 / oValue);

		case INT_VAL:
			return CDT(u.i_data / oValue);

		case REAL_VAL:
			return CDT(u.i_data / oValue);

		default:
			throw CDTTypeCastException("operator / (W_FLOAT)");
	}

// Make stupid compiler happy
return CDT(*this);
}

//
// Operator / for type CDT
//
CDT CDT::operator/(const CDT & oCDT)
{
	switch (eValueType)
	{
		case UNDEF:
			if      (oCDT.eValueType == UNDEF)    { throw CTPPLogicError("Illegal division by zero"); }
			else if (oCDT.eValueType == INT_VAL)  { return CDT(0 / oCDT.u.i_data);     }
			else if (oCDT.eValueType == REAL_VAL) { return CDT(0 / oCDT.u.d_data);     }
			else                                  { throw CDTTypeCastException("operator / (oCDT)"); }

		case INT_VAL:
			if      (oCDT.eValueType == UNDEF)    { throw CTPPLogicError("Illegal division by zero"); }
			else if (oCDT.eValueType == INT_VAL)  { return CDT(u.i_data / oCDT.u.i_data);     }
			else if (oCDT.eValueType == REAL_VAL) { return CDT(u.i_data / oCDT.u.d_data);     }
			else                                  { throw CDTTypeCastException("operator / (oCDT)"); }

		case REAL_VAL:
			if      (oCDT.eValueType == UNDEF)    { throw CTPPLogicError("Illegal division by zero"); }
			else if (oCDT.eValueType == INT_VAL)  { return CDT(u.d_data / oCDT.u.i_data);     }
			else if (oCDT.eValueType == REAL_VAL) { return CDT(u.d_data / oCDT.u.d_data);     }
			else                                  { throw CDTTypeCastException("operator / (oCDT)"); }
		default:
			throw CDTTypeCastException("operator / (CDT)");
	}

// Make stupid compiler happy
return CDT(*this);
}

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator += for type INT_64
//
CDT & CDT::operator+=(const INT_64 & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			::new (this)CDT(oValue);
			break;

		case INT_VAL:
			u.i_data += oValue;
			break;

		case REAL_VAL:
			u.d_data += oValue;
			break;

		case STRING_VAL:
			{
				Unshare();
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];
				snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "%llu", (long long) (oValue));
				u.p_data -> u.s_data -> append(szBuf);
			}
			break;

		default:
			throw CDTTypeCastException("operator +=");
	}

return *this;
}

//
// Operator += for type UINT_64
//
CDT & CDT::operator+=(const UINT_64 & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			::new (this)CDT(oValue);
			break;

		case INT_VAL:
			u.i_data += oValue;
			break;

		case REAL_VAL:
			u.d_data += oValue;
			break;

		case STRING_VAL:
			{
				Unshare();
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];
				snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "%llu", (long long) (oValue));
				u.p_data -> u.s_data -> append(szBuf);
			}
			break;

		default:
			throw CDTTypeCastException("operator +=");
	}

return *this;
}

//
// Operator += for type INT_32
//
CDT & CDT::operator+=(const INT_32 & oValue)  { return this -> operator+=(INT_64(oValue));  }

//
// Operator += for type UINT_32
//
CDT & CDT::operator+=(const UINT_32 & oValue) { return this -> operator+=(UINT_64(oValue)); }

//
// Operator += for type W_FLOAT
//
CDT & CDT::operator+=(const W_FLOAT & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			::new (this)CDT(oValue);
			break;

		case INT_VAL:
			{
				eValueType = REAL_VAL;
				W_FLOAT dTMP = u.i_data + oValue;
				u.d_data = dTMP;
			}
			break;

		case REAL_VAL:
			u.d_data += oValue;
			break;

		case STRING_VAL:
			{
				Unshare();
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];
				snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "%.*G", CTPP_FLOAT_PRECISION, oValue);
				u.p_data -> u.s_data -> append(szBuf);
			}
			break;

		default:
			throw CDTTypeCastException("operator +=");
	}

return *this;
}

//
// Operator += for type CDT
//
CDT & CDT::operator+=(const CDT & oCDT)
{
	switch (eValueType)
	{
		case UNDEF:
			eValueType = oCDT.eValueType;
			if      (eValueType == UNDEF)      { ;; }
			else if (eValueType == INT_VAL)    { u.i_data = oCDT.u.i_data; }
			else if (eValueType == REAL_VAL)   { u.d_data = oCDT.u.d_data; }
			else if (eValueType == STRING_VAL) { ::new (this) CDT(oCDT);   }
			else { throw CDTTypeCastException("operator +="); }
			break;

		case INT_VAL:
			if      (oCDT.eValueType == UNDEF)   { ;; }
			else if (oCDT.eValueType == INT_VAL) { u.i_data += oCDT.u.i_data; }
			else if (oCDT.eValueType == REAL_VAL)
			{
				eValueType = REAL_VAL;
				W_FLOAT dTMP = u.i_data + oCDT.u.d_data;
				u.d_data = dTMP;
			}
			else if (oCDT.eValueType == STRING_VAL)
			{
				const std::string sTMP = this -> GetString();
				::new (this) CDT(sTMP);
				u.p_data -> u.s_data -> append(*(oCDT.u.p_data -> u.s_data));
			}
			else { throw CDTTypeCastException("operator +="); }
			break;

		case REAL_VAL:
			if      (oCDT.eValueType == UNDEF)    { ;; }
			else if (oCDT.eValueType == INT_VAL)  { u.d_data += oCDT.u.i_data; }
			else if (oCDT.eValueType == REAL_VAL) { u.d_data += oCDT.u.d_data; }
			else if (oCDT.eValueType == STRING_VAL)
			{
				const std::string sTMP = this -> GetString();
				::new (this) CDT(sTMP);
				u.p_data -> u.s_data -> append(*(oCDT.u.p_data -> u.s_data));
			}
			else { throw CDTTypeCastException("operator +="); }
			break;

		case STRING_VAL:
			Unshare();
			if      (oCDT.eValueType == UNDEF)    { ;; }
			else if (oCDT.eValueType == INT_VAL  ||
				 oCDT.eValueType == REAL_VAL ||
				 oCDT.eValueType == STRING_VAL) { u.p_data -> u.s_data -> append(oCDT.GetString()); }
			else { throw CDTTypeCastException("operator +="); }
			break;
		default:
			throw CDTTypeCastException("operator +=");
	}

return *this;
}

//
// Operator += for type std::string
//
CDT & CDT::operator+=(const std::string & oValue)
{
	if (eValueType == UNDEF)
	{
		::new (this)CDT(oValue);
	}
	else if (eValueType == INT_VAL ||
	         eValueType == REAL_VAL)
	{
		std::string sTMP(GetString() + oValue);
		::new (this)CDT(sTMP);
	}
	else if (eValueType == STRING_VAL)
	{
		Unshare();
		u.p_data -> u.s_data -> append(oValue);
	}
	else { throw CDTTypeCastException("operator +="); }

return *this;
}

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator -= for type INT_32
//
CDT & CDT::operator-=(const INT_64 & oValue)  { return this -> operator+=(- oValue);  }

//
// Operator -= for type UINT_32
//
CDT & CDT::operator-=(const UINT_64 & oValue) { return this -> operator+=(- INT_64(oValue));  }

//
// Operator -= for type INT_32
//
CDT & CDT::operator-=(const INT_32 & oValue)  { return this -> operator+=(- oValue);  }

//
// Operator -= for type UINT_32
//
CDT & CDT::operator-=(const UINT_32 & oValue) { return this -> operator+=(- INT_32(oValue));  }

//
// Operator -= for type W_FLOAT
//
CDT & CDT::operator-=(const W_FLOAT & oValue) { return this -> operator+=(- oValue); }

//
// Operator -= for type CDT
//
CDT & CDT::operator-=(const CDT & oCDT)
{
	switch (eValueType)
	{
		case UNDEF:
			eValueType = oCDT.eValueType;
			if      (eValueType == UNDEF)    { ;; }
			else if (eValueType == INT_VAL)  { u.i_data = - oCDT.u.i_data; }
			else if (eValueType == REAL_VAL) { u.d_data = - oCDT.u.d_data; }
			else { throw CDTTypeCastException("operator -="); }
			break;

		case INT_VAL:
			if      (oCDT.eValueType == UNDEF)    { ;; }
			else if (oCDT.eValueType == INT_VAL)  { u.i_data -= oCDT.u.i_data; }
			else if (oCDT.eValueType == REAL_VAL)
			{
				eValueType = REAL_VAL;
				W_FLOAT dTMP = u.i_data - oCDT.u.d_data;
				u.d_data = dTMP;
			}
			else { throw CDTTypeCastException("operator -="); }
			break;

		case REAL_VAL:
			if      (oCDT.eValueType == UNDEF)    { ;; }
			else if (oCDT.eValueType == INT_VAL)  { u.d_data -= oCDT.u.i_data; }
			else if (oCDT.eValueType == REAL_VAL) { u.d_data -= oCDT.u.d_data; }
			else { throw CDTTypeCastException("operator -="); }
			break;

		default:
			throw CDTTypeCastException("operator -=");
	}
return *this;
}

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator *= for type INT_64
//
CDT & CDT::operator*=(const INT_64 & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			::new (this)CDT(INT_64(0));
			break;

		case INT_VAL:
			u.i_data *= oValue;
			break;

		case REAL_VAL:
			u.d_data *= oValue;
			break;

		default:
			throw CDTTypeCastException("operator *=");
	}

return *this;
}

//
// Operator *= for type UINT_64
//
CDT & CDT::operator*=(const UINT_64 & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			::new (this)CDT(UINT_64(0));
			break;

		case INT_VAL:
			u.i_data *= oValue;
			break;

		case REAL_VAL:
			u.d_data *= oValue;
			break;

		default:
			throw CDTTypeCastException("operator *=");
	}

return *this;
}

//
// Operator *= for type INT_32
//
CDT & CDT::operator*=(const INT_32 & oValue)  { return this -> operator*=(INT_64(oValue));  }

//
// Operator *= for type UINT_32
//
CDT & CDT::operator*=(const UINT_32 & oValue) { return this -> operator*=(INT_64(oValue));  }

//
// Operator *= for type W_FLOAT
//
CDT & CDT::operator*=(const W_FLOAT & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			::new (this)CDT(W_FLOAT(0.0));
			break;

		case INT_VAL:
			{
				eValueType = REAL_VAL;
				W_FLOAT dTMP = u.i_data * oValue;
				u.d_data = dTMP;
			}
			break;

		case REAL_VAL:
			u.d_data *= oValue;
			break;

		default:
			throw CDTTypeCastException("operator *=");
	}

return *this;
}

//
// Operator *= for type CDT
//
CDT & CDT::operator*=(const CDT & oCDT)
{
	switch (eValueType)
	{
		case UNDEF:
			eValueType = oCDT.eValueType;
			if      (eValueType == UNDEF)      { ;; }
			else if (eValueType == INT_VAL)    { u.i_data = 0; }
			else if (eValueType == REAL_VAL)   { u.d_data = 0; }
			else    { throw CDTTypeCastException("operator *="); }
			break;

		case INT_VAL:
			if      (oCDT.eValueType == UNDEF)    { ;; }
			else if (oCDT.eValueType == INT_VAL)  { u.i_data *= oCDT.u.i_data; }
			else if (oCDT.eValueType == REAL_VAL)
			{
				eValueType = REAL_VAL;
				W_FLOAT dTMP = u.i_data * oCDT.u.d_data;
				u.d_data = dTMP;
			}
			else    { throw CDTTypeCastException("operator *="); }
			break;

		case REAL_VAL:
			if      (oCDT.eValueType == UNDEF)    { ;; }
			else if (oCDT.eValueType == INT_VAL)  { u.d_data *= oCDT.u.i_data; }
			else if (oCDT.eValueType == REAL_VAL) { u.d_data *= oCDT.u.d_data; }
			else    { throw CDTTypeCastException("operator *="); }
			break;

		default:
			throw CDTTypeCastException("operator *=");
	}

return *this;
}

// ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator /= for type INT_64
//
CDT & CDT::operator/=(const INT_64 & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			::new (this)CDT(INT_64(0) / oValue);
			break;

		case INT_VAL:
			u.i_data /= oValue;
			break;

		case REAL_VAL:
			u.d_data /= oValue;
			break;

		default:
			throw CDTTypeCastException("operator /=");
	}

return *this;
}

//
// Operator /= for type UINT_64
//
CDT & CDT::operator/=(const UINT_64 & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			::new (this)CDT(UINT_64(0) / oValue);
			break;

		case INT_VAL:
			u.i_data /= oValue;
			break;

		case REAL_VAL:
			u.d_data /= oValue;
			break;

		default:
			throw CDTTypeCastException("operator /=");
	}

return *this;
}

//
// Operator /= for type INT_32
//
CDT & CDT::operator/=(const INT_32 & oValue)  { return this -> operator/=(INT_64(oValue));  }

//
// Operator /= for type UINT_32
//
CDT & CDT::operator/=(const UINT_32 & oValue) { return this -> operator/=(UINT_64(oValue)); }

//
// Operator /= for type W_FLOAT
//
CDT & CDT::operator/=(const W_FLOAT & oValue)
{
	switch (eValueType)
	{
		case UNDEF:
			::new (this)CDT(0.0 / oValue);
			break;

		case INT_VAL:
			{
				eValueType = REAL_VAL;
				W_FLOAT dTMP = u.i_data / oValue;
				u.d_data = dTMP;
			}
			break;

		case REAL_VAL:
			u.d_data /= oValue;
			break;

		default:
			throw CDTTypeCastException("operator /=");
	}

return *this;
}

//
// Operator /= for type CDT
//
CDT & CDT::operator/=(const CDT & oCDT)
{
	switch (eValueType)
	{
		case UNDEF:
			eValueType = oCDT.eValueType;
			if      (eValueType == UNDEF)    { throw CTPPLogicError("Illegal division by zero"); }
			else if (eValueType == INT_VAL)  { u.i_data = 0; }
			else if (eValueType == REAL_VAL) { u.d_data = 0; }
			else    { throw CDTTypeCastException("operator /="); }
			break;

		case INT_VAL:
			if      (oCDT.eValueType == UNDEF)    { throw CTPPLogicError("Illegal division by zero"); }
			else if (oCDT.eValueType == INT_VAL)  { u.i_data /= oCDT.u.i_data; }
			else if (oCDT.eValueType == REAL_VAL)
			{
				eValueType = REAL_VAL;
				W_FLOAT dTMP = u.i_data / oCDT.u.d_data;
				u.d_data = dTMP;
			}
			else    { throw CDTTypeCastException("operator /="); }
			break;

		case REAL_VAL:
			if      (oCDT.eValueType == UNDEF)    { throw CTPPLogicError("Illegal division by zero"); }
			else if (oCDT.eValueType == INT_VAL)  { u.d_data /= oCDT.u.i_data; }
			else if (oCDT.eValueType == REAL_VAL) { u.d_data /= oCDT.u.d_data; }
			else    { throw CDTTypeCastException("operator /="); }
			break;

		default:
			;;
	}
return *this;
}

// ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Append a old-fashion string to CDT
//
CDT & CDT::Concat(CCHAR_P szData, const INT_32 & iDataLength)
{
	if (eValueType == UNDEF)
	{
		::new (this)CDT(std::string(szData, iDataLength));
	}
	else if (eValueType == INT_VAL ||
	         eValueType == REAL_VAL)
	{
		std::string sTMP(GetString() + std::string(szData, iDataLength));
		::new (this)CDT(sTMP);
	}
	else if (eValueType == STRING_VAL)
	{
		Unshare();
		u.p_data -> u.s_data -> append(szData, iDataLength);
	}
	else { throw CDTTypeCastException("operator +="); }

return *this;
}

//
// Get value as W_FLOAT
//
W_FLOAT CDT::GetFloat() const
{
	switch (eValueType)
	{
		case INT_VAL:
			return W_FLOAT(u.i_data);

		case REAL_VAL:
			return u.d_data;

		default:
			return 0.0;
	}
}

//
// Get value as INT_64
//
INT_64 CDT::GetInt() const
{
	switch (eValueType)
	{
		case INT_VAL:
			return u.i_data;

		case REAL_VAL:
			return INT_64(u.d_data);

		default:
			return 0;
	}
}

//
// Get value as UINT_64
//
UINT_64 CDT::GetUInt() const
{
	switch (eValueType)
	{
		case INT_VAL:
			return u.i_data;

		case REAL_VAL:
			return INT_64(u.d_data);

		default:
			return 0;
	}
}

//
// Get value as string
//
std::string CDT::GetString(CCHAR_P szFormat) const
{
	switch (eValueType)
	{
		case INT_VAL:
			{
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];
				if (szFormat != NULL && *szFormat != '\0')
				{
					snprintf(szBuf, C_MAX_SPRINTF_LENGTH, szFormat, (long long) (u.i_data));
				}
				else
				{
					snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "%lli", (long long) (u.i_data));
				}

			return szBuf;
			}

		case REAL_VAL:
			{
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];
				if (szFormat != NULL && *szFormat != '\0')
				{
					snprintf(szBuf, C_MAX_SPRINTF_LENGTH, szFormat, u.d_data);
				}
				else
				{
					snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "%.*G", CTPP_FLOAT_PRECISION, u.d_data);
				}
			return szBuf;
			}

		case POINTER_VAL:
			{
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];
				if (szFormat != NULL && *szFormat != '\0')
				{
					snprintf(szBuf, C_MAX_SPRINTF_LENGTH, szFormat, u.pp_data);
				}
				else
				{
					snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "%p", u.pp_data);
				}
			return szBuf;
			}


		case STRING_VAL:
			return *(u.p_data -> u.s_data);

		case ARRAY_VAL:
			{
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];
				snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "ARRAY (%p)", (void *)(u.p_data -> u.s_data));
				return szBuf;
			}

		case HASH_VAL:
			{
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];
				snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "HASH (%p)", (void *)(u.p_data -> u.s_data));
				return szBuf;
			}

		default:
			return "";
	}

}

//
// Get generic pointer
//
void * CDT::GetPointer() const
{
	if (eValueType != POINTER_VAL) { return NULL; }

return u.pp_data;
}

//
// Dump CDT recursively to string
//
std::string CDT::RecursiveDump(UINT_32 iLevel) const
{
	std::string sResult;
	DumpData(iLevel, *this, sResult);
	return sResult;
}

//
// Dump to string
//
void CDT::DumpData(UINT_32 iLevel, const CDT & oData, std::string &sResult) const
{
	CHAR_8 szBuf[512 + 1];

	++iLevel;
	switch (oData.GetType())
	{
		case UNDEF:
		case INT_VAL:
		case REAL_VAL:
		case STRING_VAL:
		case POINTER_VAL:
			sResult += oData.GetString() + "\n";
			break;

		case ARRAY_VAL:
			{
				sResult += "ARRAY [\n";
				for (UINT_32 iI = 0; iI < oData.Size(); ++iI)
				{
					for (UINT_32 iJ = 0; iJ < iLevel; iJ++) { sResult += "    "; }
					snprintf(szBuf, 512, "%u", iI);
					sResult += szBuf;
					sResult += " : ";
					DumpData(iLevel, oData.GetCDT(iI), sResult);
				}
				for (UINT_32 iI = 0; iI < iLevel - 1; ++iI) { sResult += "    "; }
				sResult += "]\n";
			}
			break;

		case HASH_VAL:
			{
				sResult += "HASH {\n";
				ConstIterator itHash = oData.Begin();
				while (itHash != oData.End())
				{
					for (UINT_32 iJ = 0; iJ < iLevel; iJ++) { sResult += "    "; }
					sResult += itHash -> first;
					sResult += " => ";
					DumpData(iLevel, itHash -> second, sResult);
					++itHash;
				}
				for (UINT_32 iI = 0; iI < iLevel - 1; ++iI) { sResult += "    "; }
				sResult += "}\n";
			}
			break;

		default:
			sResult += "Invalid type";
	}
}

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator == for type INT_32
//
bool CDT::operator==(const INT_64 & oValue)  const { return (GetInt() == oValue);  }

//
// Operator == for type UINT_32
//
bool CDT::operator==(const UINT_64 & oValue) const { return (GetUInt() == oValue);  }

//
// Operator == for type INT_32
//
bool CDT::operator==(const INT_32 & oValue)  const { return (GetInt() == oValue);  }

//
// Operator == for type UINT_32
//
bool CDT::operator==(const UINT_32 & oValue) const { return (GetUInt() == oValue);  }

//
// Operator == for type W_FLOAT
//
bool CDT::operator==(const W_FLOAT & oValue) const { return (GetFloat() == oValue);  }

//
// Operator == for type CDT
//
bool CDT::operator==(const CDT & oCDT) const
{
	if      (oCDT.eValueType == STRING_VAL)    { return (GetString() == oCDT.GetString()); }
	else if (oCDT.eValueType <= REAL_VAL &&
		      eValueType <= REAL_VAL)      { return (GetFloat() == oCDT.GetFloat()); }

return false;
}

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator != for type INT_32
//
bool CDT::operator!=(const INT_64 & oValue)  const { return ! this -> operator==(INT_64(oValue));    }

//
// Operator != for type UINT_32
//
bool CDT::operator!=(const UINT_64 & oValue)  const { return ! this -> operator==(UINT_64(oValue));    }

//
// Operator != for type INT_32
//
bool CDT::operator!=(const INT_32 & oValue)  const { return ! this -> operator==(INT_32(oValue));    }

//
// Operator != for type UINT_32
//
bool CDT::operator!=(const UINT_32 & oValue)  const { return ! this -> operator==(UINT_32(oValue));    }

//
// Operator != for type W_FLOAT
//
bool CDT::operator!=(const W_FLOAT & oValue)  const { return ! this -> operator==(oValue);  }

//
// Operator != for type CDT
//
bool CDT::operator!=(const CDT & oCDT)        const { return ! this -> operator==(oCDT);        }

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator > for type INT_32
//
bool CDT::operator>(const INT_64 & oValue)  const { return (GetInt() > oValue);  }

//
// Operator > for type UINT_32
//
bool CDT::operator>(const UINT_64 & oValue) const { return (GetUInt() > oValue);  }

//
// Operator > for type INT_32
//
bool CDT::operator>(const INT_32 & oValue)  const { return (GetInt() > oValue);  }

//
// Operator > for type UINT_32
//
bool CDT::operator>(const UINT_32 & oValue) const { return (GetUInt() > oValue);  }

//
// Operator > for type W_FLOAT
//
bool CDT::operator>(const W_FLOAT & oValue) const { return (GetFloat() > oValue);  }

//
// Operator > for type CDT
//
bool CDT::operator>(const CDT & oCDT) const
{
	if      (eValueType <= REAL_VAL)  { return (GetFloat() > oCDT.GetFloat());                   }
	else if (eValueType == STRING_VAL) { return ( *(u.p_data -> u.s_data ) > oCDT.GetString() ); }
return false;
}

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator < for type INT_32
//
bool CDT::operator<(const INT_64 & oValue)  const { return (GetInt() < oValue);  }

//
// Operator < for type UINT_32
//
bool CDT::operator<(const UINT_64 & oValue) const { return (GetUInt() < oValue);  }

//
// Operator < for type INT_32
//
bool CDT::operator<(const INT_32 & oValue)  const { return (GetInt() < oValue);  }

//
// Operator < for type UINT_32
//
bool CDT::operator<(const UINT_32 & oValue) const { return (GetUInt() < oValue);  }

//
// Operator < for type W_FLOAT
//
bool CDT::operator<(const W_FLOAT & oValue) const { return (GetFloat() < oValue);  }

//
// Operator < for type CDT
//
bool CDT::operator<(const CDT & oCDT) const
{
	if      (eValueType <= REAL_VAL)  { return (GetFloat() < oCDT.GetFloat());                   }
	else if (eValueType == STRING_VAL) { return ( *(u.p_data -> u.s_data ) < oCDT.GetString() ); }
return false;
}

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator <= for type INT_32
//
bool CDT::operator<=(const INT_64 & oValue)  const { return (GetInt() <= oValue);  }

//
// Operator <= for type UINT_32
//
bool CDT::operator<=(const UINT_64 & oValue) const { return (GetUInt() <= oValue);  }

//
// Operator <= for type INT_32
//
bool CDT::operator<=(const INT_32 & oValue)  const { return (GetInt() <= oValue);  }

//
// Operator <= for type UINT_32
//
bool CDT::operator<=(const UINT_32 & oValue) const { return (GetUInt() <= oValue);  }

//
// Operator <= for type W_FLOAT
//
bool CDT::operator<=(const W_FLOAT & oValue) const { return (GetFloat() <= oValue);  }

//
// Operator <= for type CDT
//
bool CDT::operator<=(const CDT & oCDT) const
{
	if      (eValueType <= REAL_VAL)  { return (GetFloat() <= oCDT.GetFloat());                  }
	else if (eValueType == STRING_VAL) { return ( *(u.p_data -> u.s_data )<= oCDT.GetString() ); }
return false;
}

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Operator >= for type INT_32
//
bool CDT::operator>=(const INT_64 & oValue)  const { return (GetInt() >= oValue);  }

//
// Operator >= for type UINT_32
//
bool CDT::operator>=(const UINT_64 & oValue) const { return (GetUInt() >= oValue);  }

//
// Operator >= for type INT_32
//
bool CDT::operator>=(const INT_32 & oValue)  const { return (GetInt() >= oValue);  }

//
// Operator >= for type UINT_32
//
bool CDT::operator>=(const UINT_32 & oValue) const { return (GetUInt() >= oValue);  }

//
// Operator >= for type W_FLOAT
//
bool CDT::operator>=(const W_FLOAT & oValue) const { return (GetFloat() >= oValue);  }

//
// Operator >= for type CDT
//
bool CDT::operator>=(const CDT & oCDT) const
{
	if      (eValueType <= REAL_VAL)   { return (GetFloat() >= oCDT.GetFloat());                 }
	else if (eValueType == STRING_VAL) { return ( *(u.p_data -> u.s_data )>= oCDT.GetString() ); }
return false;
}

// ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// A destructor
//
CDT::~CDT() throw() { Destroy(); }

//
// Destroy object if need
//
void CDT::Destroy() throw()
{
	switch (eValueType)
	{
		// POD types
		case UNDEF:
		case INT_VAL:
		case REAL_VAL:
		case POINTER_VAL:
			;; // Nothing to do
			break;

		case STRING_VAL:
			-- (u.p_data -> refcount);
			if (u.p_data -> refcount == 0)
			{
				delete u.p_data -> u.s_data;
				delete u.p_data;
			}
			break;

		case ARRAY_VAL:
			-- (u.p_data -> refcount);
			if (u.p_data -> refcount == 0)
			{
				delete u.p_data -> u.v_data;
				delete u.p_data;
			}
			break;

		case HASH_VAL:
			-- (u.p_data -> refcount);
			if (u.p_data -> refcount == 0)
			{
				delete u.p_data -> u.m_data;
				delete u.p_data;
			}
			break;

		default:
			// R.I.P.
			{ int * pI = NULL; *pI = 0xDeadBeef; }
	}
}

//
// Unshare shareable container
//
void CDT::Unshare()
{
	if (u.p_data -> refcount != 1)
	{
		_CDT * pTMP = new _CDT();

		if      (eValueType == STRING_VAL) { pTMP -> u.s_data = new std::string(*(u.p_data -> u.s_data)); }
		else if (eValueType == ARRAY_VAL)  { pTMP -> u.v_data = new Vector(*(u.p_data -> u.v_data));      }
		else if (eValueType == HASH_VAL)   { pTMP -> u.m_data = new Map(*(u.p_data -> u.m_data));         }

		-- (u.p_data -> refcount);
		u.p_data = pTMP;
	}
}

} // namespace CTPP
// End.
