/*-
 * Copyright (c) 2004 - 2010 CTPP Team
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the CTPP Team nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      VM.cpp
 *
 * $CTPP$
 */

#include "VM.hpp"
#include "VMDebugInfo.hpp"
#include "VMOpcodes.h"
#include "VMException.hpp"
#include "VMStackException.hpp"
#include "VMInstruction.hpp"

#include <string.h>

/*
    Foreground colors
       30    Black
       31    Red
       32    Green
       33    Yellow
       34    Blue
       35    Magenta
       36    Cyan
       37    White

    Background colors
       40    Black
       41    Red
       42    Green
       43    Yellow
       44    Blue
       45    Magenta
       46    Cyan
       47    White
*/
#define BRIGHT 1
#define RED    31
#define YELLOW 33
#define GREEN  32
#define BLUE   34
#define BG_BLACK 40


#ifndef WIN32 /* Windows console is not so funky */
#define HL_CODE(x) fprintf(stderr, "%c[%d;%d;%dm", 0x1B, BRIGHT, (x), BG_BLACK)
#define HL_RST  fprintf(stderr, "%c[%dm", 0x1B, 0)
#else
    #define HL_CODE(x)
    #define HL_RST
#endif

// Aliases for registers
#define AR oRegs[0]
#define BR oRegs[1]
#define CR oRegs[2]
#define DR oRegs[3]
#define ER oRegs[4]
#define FR oRegs[5]
#define GR oRegs[6]
#define HR oRegs[7]

namespace CTPP // C++ Template Engine
{

//
// Constructor
//
VM::VM(SyscallFactory  & oISyscallFactory,
       const UINT_32   & iIMaxArgStackSize,
       const UINT_32   & iIMaxCodeStackSize,
       const UINT_32   & iIMaxSteps,
       const UINT_32   & iIDebugLevel): oSyscallFactory(oISyscallFactory),
                                        iMaxArgStackSize(iIMaxArgStackSize),
                                        iMaxCodeStackSize(iIMaxCodeStackSize),
                                        iMaxSteps(iIMaxSteps),
                                        iDebugLevel(iIDebugLevel),
                                        iMaxCalls(0),
                                        iMaxUsedCalls(0),
                                        aCallTranslationMap(NULL),
                                        oVMArgStack(iMaxArgStackSize),
                                        oVMCodeStack(iMaxCodeStackSize)
{
	;;
}

//
// Initialize virtual machine
//
INT_32 VM::Init(OutputCollector     & pIOutputCollector,
                const VMMemoryCore  & oMemoryCore)
{
	// Output data collector
	pOutputCollector = &pIOutputCollector;

	// Create syscalls translation map
	iMaxUsedCalls = oMemoryCore.syscalls.GetRecordsNum();
	// Cache
	if (iMaxUsedCalls >= iMaxCalls)
	{
		delete [] aCallTranslationMap;
		aCallTranslationMap = new SyscallHandler *[iMaxUsedCalls];
		iMaxCalls = iMaxUsedCalls;
	}
	// Initialize program
	for (UINT_32 iCallNum = 0; iCallNum < iMaxUsedCalls; ++iCallNum)
	{
		// Ugh
		UINT_32 iCallNameLength = 0;
		// NULL-terminated string, iCallNameLength will not used
		CCHAR_P sCallName = oMemoryCore.syscalls.GetData(iCallNum, iCallNameLength);

		// Get syscall
		SyscallHandler * pTMP = oSyscallFactory.GetHandlerByName(sCallName);

		// If this syscall is not present in factory, throw exception
		if (pTMP == NULL)
		{
			UINT_32 iDataSize = 0;
			CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(oMemoryCore.instructions[0].reserved).GetDescrId(), iDataSize);

			throw InvalidSyscall(sCallName, 0, oMemoryCore.instructions[0].reserved, szTMP);
		}

		// All OK
		aCallTranslationMap[iCallNum] = pTMP;

		// Initialize syscall handler
		pTMP -> PreExecuteSetup(pIOutputCollector,
		                        DR,
		                        oMemoryCore.syscalls,
		                        oMemoryCore.static_data,
		                        oMemoryCore.static_text);
	}

return 0;
}

//
// Run program
//
INT_32 VM::Run(const VMMemoryCore  & oMemoryCore,
               UINT_32             & iIP,
               CDT                 & oCDT)
{
	DR = oCDT;
	// Get code segment
	const VMInstruction * aCode = oMemoryCore.instructions;
	const UINT_32 iCodeLength   = oMemoryCore.code_size;
	UINT_32 iExecutedSteps      = 0;

	try
	{
		while (iIP < iCodeLength)
		{
			UINT_32 iOpCode = aCode[iIP].instruction;
#ifdef _DEBUG
HL_CODE(BLUE);
fprintf(stderr, "CODE 0x%08X ARG 0x%08X RES 0x%016llX | ", iOpCode, aCode[iIP].argument, (long long)(aCode[iIP].reserved));
HL_RST;
#endif
			switch(iOpCode)
			{
				// Halt program
				case HLT:
				{
#ifdef _DEBUG
HL_CODE(BLUE);
fprintf(stderr, "0x%08X HLT\n", iIP);
HL_RST;
#endif
					return 0;
				}
				break;

				// Break execution
				case BRK:
				{
#ifdef _DEBUG
HL_CODE(BLUE);
fprintf(stderr, "0x%08X BRK\n", iIP);
HL_RST;
#endif
					++iIP;
					if (iDebugLevel > 0) { return 0; }
				}
				break;

				// System call
				case SYSCALL:
				{
					const UINT_32 iCallNum    = (aCode[iIP].argument & 0xFFFF0000) >> 16;
					const UINT_32 iCallArgNum = (aCode[iIP].argument & 0x0000FFFF);
#ifdef _DEBUG
{
UINT_32 iCallNameLength = 0;
// NULL-terminated string, iCallNameLength not need to use
CCHAR_P sCallName = oMemoryCore.syscalls.GetData(iCallNum, iCallNameLength);
HL_CODE(YELLOW);
fprintf(stderr, "0x%08X SYSCALL   0x%08X %s(ARGS: %d)\n", iIP, iCallNum, sCallName, iCallArgNum);
HL_RST;
}
#endif
					// Check call number
					if (iCallNum > iMaxUsedCalls)
					{
						UINT_32 iDataSize = 0;
						CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
						throw InvalidSyscall("*** CORRUPTED ***", iIP, aCode[iIP].reserved, szTMP);
					}

					CDT oResult(CDT::UNDEF);
					// Invoke handler
					if (aCallTranslationMap[iCallNum] -> Handler(oVMArgStack.GetStackFrame(), iCallArgNum, oResult) != 0)
					{
						UINT_32 iDataSize = 0;
						CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
						throw InvalidSyscall("*** Internal syscall error ***", iIP, aCode[iIP].reserved, szTMP);
					}

					// Clear stack
					oVMArgStack.ClearStack(iCallArgNum);

					// Store execution result into stack
					oVMArgStack.PushElement(oResult);

					++iIP;
				}
				break;

				// Call procedure
				case CALL:
				{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X CALL      0x%08X\n", iIP, aCode[iIP].argument);
#endif
					if (iExecutedSteps >= iMaxSteps)
					{
						UINT_32 iDataSize = 0;
						CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
						throw ExecutionLimitReached(iIP, aCode[iIP].reserved, szTMP);
					}

					const UINT_32 iNewIP = aCode[iIP].argument;
					if (iNewIP >= iCodeLength)
					{
						UINT_32 iDataSize = 0;
						CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
						throw CodeSegmentOverrun(iIP, aCode[iIP].reserved, szTMP);
					}

					oVMCodeStack.PushAddress(iIP + 1);
					iIP = iNewIP;
				}
				break;

				// Call procedure by name
				case CALLNAME:
				{
					// Get call name
					UINT_32 iDataSize = 0;
					CCHAR_P szCallName = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);
					UINT_32 iNewIP = UINT_32(oMemoryCore.calls_table.Get(szCallName, iDataSize));
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X CALLNAME  `%s` -> %d\n", iIP, szCallName, iNewIP);
HL_RST;
#endif
					if (iNewIP == (UINT_32)-1)
					{
						CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
						throw InvalidCall(iIP, aCode[iIP].reserved, szCallName, szTMP);
					}

					if (iNewIP >= iCodeLength)
					{
						CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
						throw CodeSegmentOverrun(iIP, aCode[iIP].reserved, szTMP);
					}

					oVMCodeStack.PushAddress(iIP + 1);
					iIP = iNewIP;
				}
				break;

				// Return from procedure
				case RET:
				{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X RET       %d\n", iIP, aCode[iIP].argument);
HL_RST;
#endif
					if (iExecutedSteps >= iMaxSteps)
					{
						UINT_32 iDataSize = 0;
						CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
						throw ExecutionLimitReached(iIP, aCode[iIP].reserved, szTMP);
					}

					// Clear stack
					oVMArgStack.ClearStack(aCode[iIP].argument);
					// Return
					iIP = oVMCodeStack.PopAddress();
				}
				break;

				// Unconditional jump
				case JMP:
				{
#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "0x%08X JMP       0x%08X\n", iIP, aCode[iIP].argument);
HL_RST;
#endif
					if (iExecutedSteps >= iMaxSteps)
					{
						UINT_32 iDataSize = 0;
						CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
						throw ExecutionLimitReached(iIP, aCode[iIP].reserved, szTMP);
					}

					const UINT_32 iNewIP = aCode[iIP].argument;
					if (iNewIP >= iCodeLength)
					{
						UINT_32 iDataSize = 0;
						CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
						throw CodeSegmentOverrun(iIP, aCode[iIP].reserved, szTMP);
					}

					iIP = iNewIP;
				}
				break;

				// Pop value from stack
				case POP:
				{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X POP       %d\n", iIP, aCode[iIP].argument);
HL_RST;
#endif
					oVMArgStack.ClearStack(aCode[iIP].argument);
					++iIP;
				}
				break;
				// No operation
				case NOP:
				{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X NOP\n", iIP);
HL_RST;
#endif
					++iIP;
				}
				break;

				// All other operations
				default:
				{
					UINT_32 iHiOpcode   = iOpCode & 0xFFFF0000;
					UINT_32 iLoHiOpcode = iOpCode & 0x0000FF00;
					UINT_32 iLoLoOpcode = iOpCode & 0x000000FF;

					switch (iHiOpcode)
					{
						// Add values
						case ADD:
						{
							if (iLoHiOpcode == ARG_DST_STACK && iLoLoOpcode == ARG_SRC_STACK)
							{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X ADD       ('%s', '%s')\n", iIP, oVMArgStack.GetTopElement(1).GetString().c_str(), oVMArgStack.GetTopElement(0).GetString().c_str());
HL_RST;
#endif
								oVMArgStack.GetTopElement(1) = oVMArgStack.GetTopElement(1) + oVMArgStack.GetTopElement(0);
								oVMArgStack.ClearStack(1);
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Substract value
						case SUB:
						{
							if (iLoHiOpcode == ARG_DST_STACK && iLoLoOpcode == ARG_SRC_STACK)
							{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X SUB       ('%s', '%s')\n", iIP, oVMArgStack.GetTopElement(1).GetString().c_str(), oVMArgStack.GetTopElement(0).GetString().c_str());
HL_RST;
#endif
								oVMArgStack.GetTopElement(1) -= oVMArgStack.GetTopElement(0);
								oVMArgStack.ClearStack(1);
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Multiplicate values
						case MUL:
						{
							if (iLoHiOpcode == ARG_DST_STACK && iLoLoOpcode == ARG_SRC_STACK)
							{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X MUL       ('%s', '%s')\n", iIP, oVMArgStack.GetTopElement(1).GetString().c_str(), oVMArgStack.GetTopElement(0).GetString().c_str());
HL_RST;
#endif
								oVMArgStack.GetTopElement(1) *= oVMArgStack.GetTopElement(0);
								oVMArgStack.ClearStack(1);
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Divide values
						case DIV:
						{
							if (iLoHiOpcode == ARG_DST_STACK && iLoLoOpcode == ARG_SRC_STACK)
							{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X DIV       ('%s', '%s')\n", iIP, oVMArgStack.GetTopElement(1).GetString().c_str(), oVMArgStack.GetTopElement(0).GetString().c_str());
HL_RST;
#endif
								oVMArgStack.GetTopElement(1) /= oVMArgStack.GetTopElement(0);
								oVMArgStack.ClearStack(1);
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Divide values
						case IDIV:
						{
							if (iLoHiOpcode == ARG_DST_STACK && iLoLoOpcode == ARG_SRC_STACK)
							{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X IDIV      ('%s', '%s')\n", iIP, oVMArgStack.GetTopElement(1).GetString().c_str(), oVMArgStack.GetTopElement(0).GetString().c_str());
HL_RST;
#endif
								INT_64 iFirst  = oVMArgStack.GetTopElement(1).GetInt();
								INT_64 iSecond = oVMArgStack.GetTopElement(0).GetInt();

								if (iSecond == 0)
								{
									UINT_32 iDataSize = 0;
									CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
									throw ZeroDivision(iIP, aCode[iIP].reserved, szTMP);
								}

								oVMArgStack.GetTopElement(1) = iFirst / iSecond;
								oVMArgStack.ClearStack(1);
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Divide values
						case MOD:
						{
							if (iLoHiOpcode == ARG_DST_STACK && iLoLoOpcode == ARG_SRC_STACK)
							{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X MOD       ('%s', '%s')\n", iIP, oVMArgStack.GetTopElement(1).GetString().c_str(), oVMArgStack.GetTopElement(0).GetString().c_str());
HL_RST;
#endif
								INT_64 iFirst  = oVMArgStack.GetTopElement(1).GetInt();
								INT_64 iSecond = oVMArgStack.GetTopElement(0).GetInt();

								oVMArgStack.GetTopElement(1) = iFirst % iSecond;
								oVMArgStack.ClearStack(1);
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Negate value
						case NEG:
						{
							if (iLoLoOpcode <= ARG_SRC_LASTREG)
							{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X NEG       %cR (%d)\n", iIP, CHAR_8(iLoLoOpcode + 'A'), INT_32(oRegs[iLoLoOpcode].GetInt()));
HL_RST;
#endif
								CDT & oTMP = oRegs[iLoLoOpcode];
								if (oTMP.GetType() <= CDT::REAL_VAL)
								{
									oTMP = 0 - oTMP;
								}
							}
							else if (iLoLoOpcode <= ARG_SRC_STACK)
							{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X NEG       STACK[%d](%s)\n", iIP, aCode[iIP].argument, oVMArgStack.GetTopElement(aCode[iIP].argument).GetString().c_str());
HL_RST;
#endif
								CDT & oTMP = oVMArgStack.GetTopElement(aCode[iIP].argument);
								if (oTMP.GetType() <= CDT::REAL_VAL)
								{
									oTMP = 0 - oTMP;
								}

							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Logical negation
						case NOT:
						{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X NOT       ", iIP);
#endif
							if (iLoLoOpcode == ARG_SRC_STACK)
							{
#ifdef _DEBUG
fprintf(stderr, "STACK[%d](%c)\n", aCode[iIP].argument, oVMArgStack.GetTopElement(aCode[iIP].argument).Defined() ? 't':'f');
HL_RST;
#endif
								CDT & oTMP = oVMArgStack.GetTopElement(aCode[iIP].argument);

								if (oTMP.Defined()) { oTMP = CDT(CDT::UNDEF); }
								// Undefined
								else                { oTMP = 1;               }
							}
							else if (iLoLoOpcode <= ARG_SRC_LASTREG)
							{
#ifdef _DEBUG
fprintf(stderr, "%cR (`%s`)\n", CHAR_8(iLoLoOpcode + 'A'), oRegs[iLoLoOpcode].GetString().c_str());
HL_RST;
#endif
								// Defined
								if (oRegs[iLoLoOpcode].Defined()) { oRegs[iLoLoOpcode] = CDT(CDT::UNDEF); }
								// Undefined
								else                              { oRegs[iLoLoOpcode] = 1;               }
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Increment register
						case INC:
						{
							if (iLoLoOpcode <= ARG_SRC_LASTREG)
							{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X INC       %cR (%d)\n", iIP, CHAR_8(iLoLoOpcode + 'A'), INT_32(oRegs[iLoLoOpcode].GetInt()));
HL_RST;
#endif
								++oRegs[iLoLoOpcode];
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Decrement register
						case DEC:
						{
							if (iLoLoOpcode <= ARG_SRC_LASTREG)
							{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X DEC       %cR (%d)\n",  iIP, CHAR_8(iLoLoOpcode + 'A'), INT_32(oRegs[iLoLoOpcode].GetInt()));
HL_RST;
#endif
								--oRegs[iLoLoOpcode];
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Compare
						case CMP:
						{
							if (iLoHiOpcode <= ARG_DST_LASTREG)
							{
								UINT_32 iDstReg = (iLoHiOpcode >> 8);
								// Register-to-register
								if (iLoLoOpcode <= ARG_SRC_LASTREG)
								{
#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "0x%08X CMP       %cR, %cR", iIP, CHAR_8(iLoLoOpcode + 'A'), CHAR_8(iDstReg + 'A'));
HL_RST;
#endif
									CDT & oFirst  = oRegs[iLoLoOpcode];
									CDT & oSecond = oRegs[iDstReg];

									// Compare as values
									if (oFirst.GetType() <= CDT::REAL_VAL && oSecond.GetType() <= CDT::REAL_VAL)
									{
										W_FLOAT dTMP = oFirst.GetFloat() - oSecond.GetFloat();
										if      (dTMP < 0.0) { iFlags = FL_LT | FL_NE; }
										else if (dTMP > 0.0) { iFlags = FL_GT | FL_NE; }
										else                 { iFlags = FL_EQ; }
#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "(`%f`, `%f`)\n", oFirst.GetFloat(), oSecond.GetFloat());
HL_RST;
#endif
										if (oFirst.GetType() == CDT::INT_VAL)
										{
											UINT_64 iTMP = oFirst.GetInt();
											if ((iTMP % 2) == 1) { iFlags |= FL_ODD; }
										}
									}
									else
									{
										CCHAR_P sAR = oFirst.GetString().c_str();
										CCHAR_P sBR = oSecond.GetString().c_str();
										INT_32 iResult = strcmp(sAR, sBR);
#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "(`%s`, `%s`)\n", sAR, sBR);
HL_RST;
#endif
										if      (iResult < 0.0) { iFlags = FL_LT | FL_NE; }
										else if (iResult > 0.0) { iFlags = FL_GT | FL_NE; }
										else                    { iFlags = FL_EQ; }
									}
								}
							}
							else if (iLoHiOpcode == ARG_DST_STACK && iLoLoOpcode == ARG_SRC_STACK)
							{
#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "0x%08X CMP       STACK[%d], STACK[%d]", iIP, 0, 1);
HL_RST;
#endif
								CDT & oFirst  = oVMArgStack.GetTopElement(1);
								CDT & oSecond = oVMArgStack.GetTopElement(0);

								// Compare as values
								if (oFirst.GetType() <= CDT::REAL_VAL && oSecond.GetType() <= CDT::REAL_VAL)
								{
									W_FLOAT dTMP = oFirst.GetFloat() - oSecond.GetFloat();
									if      (dTMP < 0.0) { iFlags = FL_LT | FL_NE; }
									else if (dTMP > 0.0) { iFlags = FL_GT | FL_NE; }
									else                 { iFlags = FL_EQ; }
#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "(`%f`, `%f`)\n", oFirst.GetFloat(), oSecond.GetFloat());
HL_RST;
#endif
									if (oFirst.GetType() == CDT::INT_VAL)
									{
										UINT_64 iTMP = oFirst.GetInt();
										if ((iTMP % 2) == 1) { iFlags |= FL_ODD; }
									}
								}
								else
								{
									CCHAR_P sAR = oFirst.GetString().c_str();
									CCHAR_P sBR = oSecond.GetString().c_str();
									INT_32 iResult = strcmp(sAR, sBR);
#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "(`%s`, `%s`)\n", sAR, sBR);
HL_RST;
#endif
									if      (iResult < 0) { iFlags = FL_LT | FL_NE; }
									else if (iResult > 0) { iFlags = FL_GT | FL_NE; }
									else                  { iFlags = FL_EQ;         }
								}
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Conditional jump
						case JXX:
						{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X JXX       0x%08X (0x%08X <-> 0x%08X)\n", iIP, aCode[iIP].argument, iLoLoOpcode, iFlags);
#endif
							if (iExecutedSteps >= iMaxSteps)
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw ExecutionLimitReached(iIP, aCode[iIP].reserved, szTMP);
							}

							const UINT_32 iNewIP = aCode[iIP].argument;
							if (iNewIP >= iCodeLength)
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw CodeSegmentOverrun(iIP, aCode[iIP].reserved, szTMP);
							}

							if ((iLoLoOpcode & iFlags) != 0) { iIP = aCode[iIP].argument; }
							else { ++iIP; }
						}
						break;

						// Check existence of operand
						case EXIST:
						{
#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "0x%08X EXIST     ", iIP);
#endif
							if (iLoLoOpcode == ARG_SRC_STACK)
							{
#ifdef _DEBUG
fprintf(stderr, "STACK[%d](%c)\n", aCode[iIP].argument, oVMArgStack.GetTopElement(aCode[iIP].argument).Defined() ? 't':'f');
HL_RST;
#endif
								// Defined
								if (oVMArgStack.GetTopElement(aCode[iIP].argument).Defined()) { iFlags = FL_EQ; }
								// Undefined
								else { iFlags = FL_NE; }
							}
							else if (iLoLoOpcode <= ARG_SRC_LASTREG)
							{
#ifdef _DEBUG
fprintf(stderr, "%cR (`%s`)\n", CHAR_8(iLoLoOpcode + 'A'), oRegs[iLoLoOpcode].GetString().c_str());
HL_RST;
#endif
								// Defined
								if (oRegs[iLoLoOpcode].Defined()) { iFlags = FL_EQ; }
								// Undefined
								else                              { iFlags = FL_NE; }
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Move ONLY to registers from registers or from stack
						case MOV:
						{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X MOV       ", iIP);
#endif
							if (iLoHiOpcode <= ARG_DST_LASTREG)
							{
								UINT_32 iDstReg = (iLoHiOpcode >> 8);
#ifdef _DEBUG
fprintf(stderr, "%cR, ", CHAR_8(iDstReg + 'A'));
#endif
								// Register-to-register
								if      (iLoLoOpcode <= ARG_SRC_LASTREG)
								{
#ifdef _DEBUG
fprintf(stderr, "%cR == `%s`\n", CHAR_8(iLoLoOpcode + 'A'), oRegs[iLoLoOpcode].GetString().c_str());
HL_RST;
#endif
									oRegs[iDstReg] = oRegs[iLoLoOpcode];
								}
								// Stack element-to-register
								else if (iLoLoOpcode == ARG_SRC_STACK)
								{
#ifdef _DEBUG
fprintf(stderr, "STACK[%d] = `%s`\n", aCode[iIP].argument, oVMArgStack.GetTopElement(aCode[iIP].argument).GetString().c_str());
HL_RST;
#endif
									oRegs[iDstReg] = oVMArgStack.GetTopElement(aCode[iIP].argument);
								}
								// Integer-to-register
								else if (iLoLoOpcode == ARG_SRC_INT)
								{
#ifdef _DEBUG
fprintf(stderr, "INT POS: %d (VAL: %d)\n", aCode[iIP].argument, INT_32(oMemoryCore.static_data.GetInt(aCode[iIP].argument)));
HL_RST;
#endif
									oRegs[iDstReg] = oMemoryCore.static_data.GetInt(aCode[iIP].argument);
								}
								// Float-to-register
								else if (iLoLoOpcode == ARG_SRC_FLOAT)
								{
#ifdef _DEBUG
fprintf(stderr, "FLOAT POS: %d (VAL: %f)\n", aCode[iIP].argument, oMemoryCore.static_data.GetFloat(aCode[iIP].argument));
HL_RST;
#endif
									oRegs[iDstReg] = oMemoryCore.static_data.GetFloat(aCode[iIP].argument);
								}
								// String-to-register
								else if (iLoLoOpcode == ARG_SRC_STR)
								{
									UINT_32 iDataSize = 0;
									CCHAR_P szTMP = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);
#ifdef _DEBUG
fprintf(stderr, "STRING POS: %d (VAL: `%s`)\n", aCode[iIP].argument, szTMP);
HL_RST;
#endif
									oRegs[iDstReg] = std::string(szTMP, iDataSize);
								}
								else
								{
									UINT_32 iDataSize = 0;
									CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
									throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
								}
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Move indirect ARRAY to register
						case MOVIINT:
						{
							UINT_32 iDstReg = (iLoHiOpcode >> 8);
							oRegs[iDstReg] = oRegs[iLoLoOpcode].GetCDT(aCode[iIP].argument);
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X MOVIINT   %cR, %cR[%d] (`%s`)", iIP, CHAR_8(iDstReg + 'A'), CHAR_8(iLoLoOpcode + 'A'), aCode[iIP].argument, oRegs[iDstReg].GetString().c_str());
HL_RST;
#endif
						}
						break;

						// Move indirect HASH to register
						case MOVISTR:
						{
							if (iLoLoOpcode > ARG_SRC_LASTREG)
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							UINT_32 iDstReg = (iLoHiOpcode >> 8);

							if (iDstReg > ARG_SRC_LASTREG)
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							UINT_32 iDataSize = 0;
							CCHAR_P szTMP = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);
//fprintf(stderr, "DUMP: %s\n", oRegs[iLoLoOpcode].RecursiveDump().c_str());
							INT_32 iCDTExist = 0;
							oRegs[iDstReg] = oRegs[iLoLoOpcode].GetExistedCDT(std::string(szTMP, iDataSize), iCDTExist);

							// Found
							if (iCDTExist == 0) { iFlags = FL_EQ; }
							// Not Found
							else                { iFlags = FL_NE; }
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X MOVISTR   %cR, %cR[%s] (`%s`)\n", iIP, CHAR_8(iDstReg + 'A'), CHAR_8(iLoLoOpcode + 'A'), szTMP, oRegs[iDstReg].GetString().c_str());
HL_RST;
#endif

							++iIP;
						}
						break;

						// Move register to indirect ARRAY
						case IMOVINT:
						{
							UINT_32 iDstReg = (iLoHiOpcode >> 8);
							oRegs[iLoLoOpcode][aCode[iIP].argument] = oRegs[iDstReg];
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X IMOVINT   %cR[%d], %cR[%d] (`%s`)\n", iIP, CHAR_8(iDstReg + 'A'), aCode[iIP].argument, CHAR_8(iLoLoOpcode + 'A'), aCode[iIP].argument, oRegs[iDstReg].GetString().c_str());
HL_RST;
#endif
							++iIP;
						}

						// Move register to indirect HASH
						case IMOVSTR:
						{
							UINT_32 iDstReg = (iLoHiOpcode >> 8);
							UINT_32 iDataSize = 0;
							CCHAR_P szTMP = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);
							oRegs[iLoLoOpcode][std::string(szTMP, iDataSize)] = oRegs[iDstReg];
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X IMOVSTR   %cR[%s], %cR (`%s`)", iIP, CHAR_8(iDstReg + 'A'), szTMP, CHAR_8(iLoLoOpcode + 'A'), oRegs[iDstReg].GetString().c_str());
HL_RST;
#endif
							++iIP;
						}
						break;

						// Get size of array
						case MOVSIZE:
						{
							if (iLoHiOpcode <= ARG_DST_LASTREG)
							{
								UINT_32 iDstReg = (iLoHiOpcode >> 8);
								// Register-to-register
								if (iLoLoOpcode <= ARG_SRC_LASTREG)
								{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X MOVSIZE   %cR, %cR.Size() (%d)\n", iIP, CHAR_8(iDstReg + 'A'), CHAR_8(iLoLoOpcode + 'A'), UINT_32(oRegs[iLoLoOpcode].Size()));
HL_RST;
#endif
									oRegs[iDstReg] = oRegs[iLoLoOpcode].Size();
								}
								else
								{
									UINT_32 iDataSize = 0;
									CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
									throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
								}
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Move indirect
						case MOVIREG:
						{
							if (iLoHiOpcode <= ARG_DST_LASTREG)
							{
								UINT_32 iDstReg = (iLoHiOpcode >> 8);
								// Register-to-register
								if (iLoLoOpcode <= ARG_SRC_LASTREG)
								{
									UINT_32 iArgNum = aCode[iIP].argument;
									if (iArgNum > ARG_SRC_LASTREG)
									{
										UINT_32 iDataSize = 0;
										CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
										throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
									}
#ifdef _DEBUG
HL_CODE(GREEN);
fprintf(stderr, "0x%08X MOVIREG   %cR, %cR[%cR] (%d): %s\n", iIP, CHAR_8(iDstReg + 'A'), CHAR_8(iLoLoOpcode + 'A'), CHAR_8(iArgNum + 'A'), INT_32(oRegs[iArgNum].GetInt()), oRegs[iLoLoOpcode][UINT_32(oRegs[iArgNum].GetInt())].GetString().c_str());
HL_RST;
#endif
									oRegs[iDstReg] = oRegs[iLoLoOpcode][UINT_32(oRegs[iArgNum].GetInt())];
//fprintf(stderr, "DUMP: %s\n", oRegs[iDstReg].RecursiveDump().c_str());
								}
								else
								{
									UINT_32 iDataSize = 0;
									CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
									throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
								}
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Push value to stack
						case PUSH:
						{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X PUSH      ", iIP);
#endif
							// From static text segment to stack
							if (iLoLoOpcode == ARG_SRC_STR)
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);
#ifdef _DEBUG
fprintf(stderr, "STRING POS: %d (VAL: `%s`)\n", aCode[iIP].argument, szTMP);
HL_RST;
#endif
								oVMArgStack.PushElement(std::string(szTMP, iDataSize));
							}
							// From static data segment to stack (integer value)
							else if (iLoLoOpcode == ARG_SRC_INT)
							{
#ifdef _DEBUG
fprintf(stderr, "INT POS: %d (VAL: %d)\n", aCode[iIP].argument, INT_32(oMemoryCore.static_data.GetInt(aCode[iIP].argument)));
HL_RST;
#endif
								oVMArgStack.PushElement(oMemoryCore.static_data.GetInt(aCode[iIP].argument));
							}
							// From static data segment to stack (float value)
							else if (iLoLoOpcode == ARG_SRC_FLOAT)
							{
#ifdef _DEBUG
fprintf(stderr, "FLOAT POS: %d (VAL: %f)\n", aCode[iIP].argument, oMemoryCore.static_data.GetFloat(aCode[iIP].argument));
HL_RST;
#endif
								oVMArgStack.PushElement(oMemoryCore.static_data.GetFloat(aCode[iIP].argument));
							}
							// From register to stack
							else if (iLoLoOpcode <= ARG_SRC_LASTREG)
							{
								oVMArgStack.PushElement(oRegs[iLoLoOpcode]);
#ifdef _DEBUG
fprintf(stderr, "%cR\n", CHAR_8(iLoLoOpcode + 'A'));
HL_RST;
#endif
							}
							// Indirect operations works ONLY with registers AR - HR
							else if (iLoLoOpcode == ARG_SRC_IND_VAL)
							{
								if (iLoHiOpcode <= ARG_DST_LASTREG)
								{
#ifdef _DEBUG
fprintf(stderr, "%cR[%d] (`%s`)\n", CHAR_8((iLoHiOpcode >> 8) + 'A'), aCode[iIP].argument, oRegs[iLoHiOpcode >> 8].GetCDT(aCode[iIP].argument).GetString().c_str());
HL_RST;
#endif
									oVMArgStack.PushElement(oRegs[iLoHiOpcode >> 8].GetCDT(aCode[iIP].argument));
								}
								else
								{
									UINT_32 iDataSize = 0;
									CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
									throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
								}
							}
							else if (iLoLoOpcode == ARG_SRC_IND_STR)
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);

								if (iLoHiOpcode <= ARG_DST_LASTREG)
								{
									INT_32 iCDTExist = 0;
#ifdef _DEBUG
fprintf(stderr, "%cR[\"%s\"] ", CHAR_8((iLoHiOpcode >> 8) + 'A'), szTMP);
fprintf(stderr, "(`%s`)\n", oRegs[iLoHiOpcode >> 8].GetExistedCDT(std::string(szTMP, iDataSize), iCDTExist).GetString().c_str());
HL_RST;
#endif
									oVMArgStack.PushElement(oRegs[iLoHiOpcode >> 8].GetExistedCDT(std::string(szTMP, iDataSize), iCDTExist));

									// Found
									if (iCDTExist == 0) { iFlags = FL_EQ; }
									// Not Found
									else                { iFlags = FL_NE; }
								}
								else
								{
									UINT_32 iDataSize = 0;
									CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
									throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
								}
							}
							else if (iLoLoOpcode == ARG_SRC_STACK)
							{
#ifdef _DEBUG
fprintf(stderr, "STACK[%d](%s)\n", aCode[iIP].argument, oVMArgStack.GetTopElement(aCode[iIP].argument).GetString().c_str());
HL_RST;
#endif
								oVMArgStack.PushElement(oVMArgStack.GetTopElement(aCode[iIP].argument));
							}
							// Illegal operation
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Replace variable in stack or register
						case REPLACE:
						{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X REPLACE   ", iIP);
#endif
							// From static text segment to stack
							if (iLoLoOpcode == ARG_SRC_IND_STR)
							{
								if (iLoHiOpcode <= ARG_DST_LASTREG)
								{
									UINT_32 iDataSize = 0;
									CCHAR_P szTMP = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);
									UINT_32 iDstReg = (iLoHiOpcode >> 8);
#ifdef _DEBUG
fprintf(stderr, "%cR[\"%s\"] (`%s`)\n", CHAR_8(iDstReg + 'A'), szTMP, oRegs[iDstReg].GetCDT(std::string(szTMP, iDataSize)).GetString().c_str());
HL_RST;
#endif
									oVMArgStack.GetTopElement(0) = oRegs[iDstReg].GetCDT(std::string(szTMP, iDataSize));
								}
								else if (iLoHiOpcode == ARG_DST_STACK)
								{
									UINT_32 iDataSize = 0;
									CCHAR_P szTMP = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);
									CDT & oTopStack  = oVMArgStack.GetTopElement(0);
									CDT oTMP = oTopStack.GetCDT(std::string(szTMP, iDataSize));
#ifdef _DEBUG
fprintf(stderr, "TOP STACK[\"%s\"] (`%s`)\n", szTMP, oTMP.GetString().c_str());
HL_RST;
#endif


									oTopStack = oTMP;
								}
							}
							++iIP;
						}
						break;

						// Clear register
						case CLEAR:
						{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X CLEAR     ", iIP);
#endif

							if (iLoHiOpcode <= ARG_DST_LASTREG)
							{
#ifdef _DEBUG
fprintf(stderr, "%cR\n", CHAR_8((iLoHiOpcode >> 8) + 'A'));
HL_RST;
#endif
								UINT_32 iDstReg = (iLoHiOpcode >> 8);
								oRegs[iDstReg] = CDT();
							}
							else if (iLoHiOpcode == ARG_DST_STACK)
							{
#ifdef _DEBUG
fprintf(stderr, "STACK[%d](%s)\n", aCode[iIP].argument, oVMArgStack.GetTopElement(aCode[iIP].argument).GetString().c_str());
HL_RST;
#endif
								oVMArgStack.GetTopElement(aCode[iIP].argument) = CDT(CDT::UNDEF);
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						// Direct output
						case OUTPUT:
						{
#ifdef _DEBUG
HL_CODE(YELLOW);
fprintf(stderr, "0x%08X OUTPUT    ", iIP);
#endif
							if (iLoLoOpcode == ARG_SRC_STACK)
							{
#ifdef _DEBUG
fprintf(stderr, "STACK[%d](`%s`)\n", aCode[iIP].argument, oVMArgStack.GetTopElement(aCode[iIP].argument).GetString().c_str());
HL_RST;
#endif
								std::string sTMP = oVMArgStack.GetTopElement(0).GetString();
								pOutputCollector -> Collect(sTMP.c_str(), sTMP.size());
								oVMArgStack.ClearStack(1);
							}
							// From static text segment
							else if (iLoLoOpcode == ARG_SRC_STR)
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);
#ifdef _DEBUG
fprintf(stderr, "STRING POS: %d (VAL: `%s`)\n", aCode[iIP].argument, szTMP);
HL_RST;
#endif
								pOutputCollector -> Collect(szTMP, iDataSize);
							}
							// From static data segment (integer value)
							else if (iLoLoOpcode == ARG_SRC_INT)
							{
#ifdef _DEBUG
fprintf(stderr, "INT POS: %d (VAL: %d)\n", aCode[iIP].argument, INT_32(oMemoryCore.static_data.GetInt(aCode[iIP].argument)));
HL_RST;
#endif
								CDT oTMP(oMemoryCore.static_data.GetInt(aCode[iIP].argument));
								std::string sTMP = oTMP.GetString();
								pOutputCollector -> Collect(sTMP.c_str(), sTMP.size());
							}
							// From static data segment (float value)
							else if (iLoLoOpcode == ARG_SRC_FLOAT)
							{
#ifdef _DEBUG
fprintf(stderr, "FLOAT POS: %d (VAL: %f)\n", aCode[iIP].argument, oMemoryCore.static_data.GetFloat(aCode[iIP].argument));
HL_RST;
#endif
								CDT oTMP(oMemoryCore.static_data.GetFloat(aCode[iIP].argument));
								std::string sTMP = oTMP.GetString();
								pOutputCollector -> Collect(sTMP.c_str(), sTMP.size());
							}
							// From register
							else if (iLoLoOpcode <= ARG_SRC_LASTREG)
							{
#ifdef _DEBUG
fprintf(stderr, "%cR\n", CHAR_8(iLoLoOpcode + 'A'));
HL_RST;
#endif
								std::string sTMP = oRegs[iLoLoOpcode].GetString();
								pOutputCollector -> Collect(sTMP.c_str(), sTMP.size());

							}
							// Indirect operations works ONLY with registers AR - HR
							else if (iLoLoOpcode == ARG_SRC_IND_VAL)
							{
#ifdef _DEBUG
fprintf(stderr, "%cR[%d] (`%s`)\n", CHAR_8((iLoHiOpcode >> 8) + 'A'), aCode[iIP].argument, oRegs[iLoHiOpcode >> 8].GetCDT(aCode[iIP].argument).GetString().c_str());
HL_RST;
#endif
								if (iLoHiOpcode <= ARG_DST_LASTREG)
								{
									std::string sTMP = oRegs[iLoHiOpcode >> 8].GetCDT(aCode[iIP].argument).GetString();
									pOutputCollector -> Collect(sTMP.c_str(), sTMP.size());
								}
								else
								{
									UINT_32 iDataSize = 0;
									CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
									throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
								}
							}
							else if (iLoLoOpcode == ARG_SRC_IND_STR)
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(aCode[iIP].argument, iDataSize);
#ifdef _DEBUG
fprintf(stderr, "%cR[\"%s\"] (`%s`)\n", CHAR_8((iLoHiOpcode >> 8) + 'A'), szTMP, oRegs[iLoHiOpcode >> 8].GetCDT(std::string(szTMP, iDataSize)).GetString().c_str());
HL_RST;
#endif
								if (iLoHiOpcode <= ARG_DST_LASTREG)
								{
									INT_32 iCDTExist = 0;
									std::string sTMP = oRegs[iLoHiOpcode >> 8].GetExistedCDT(std::string(szTMP, iDataSize), iCDTExist).GetString();
									pOutputCollector -> Collect(sTMP.c_str(), sTMP.size());
								}
								else
								{
									UINT_32 iDataSize = 0;
									CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
									throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
								}
							}
							// Illegal operation
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}

							++iIP;
						}
						break;

						case CALL:
						{
#ifdef _DEBUG
HL_CODE(RED);
fprintf(stderr, "0x%08X CALLIND   ", iIP);
#endif
							if (iLoHiOpcode == 0x0200)
							{
								std::string sCallName;
								if (iLoLoOpcode <= ARG_SRC_HR)
								{
									sCallName = oRegs[iLoLoOpcode].GetString();
#ifdef _DEBUG
fprintf(stderr, "%cR (`%s`)\n", CHAR_8(iLoLoOpcode + 'A'), oRegs[iLoLoOpcode].GetString().c_str());
HL_RST;
#endif
								}
								else if (iLoLoOpcode == ARG_SRC_STACK)
								{
									sCallName = oVMArgStack.GetTopElement(0).GetString();
#ifdef _DEBUG
fprintf(stderr, "STACK[%d] (`%s`)\n", 0, oVMArgStack.GetTopElement(0).GetString().c_str());
HL_RST;
#endif

									oVMArgStack.ClearStack(1);
								}
								UINT_32 iNewIP = UINT_32(oMemoryCore.calls_table.Get(sCallName.c_str(), sCallName.size()));
								if (iNewIP == (UINT_32)-1)
								{
									UINT_32 iDataSize = 0;
									CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
									throw InvalidCall(iIP, aCode[iIP].reserved, sCallName.c_str(), szTMP);
								}

								if (iNewIP >= iCodeLength)
								{
									UINT_32 iDataSize = 0;
									CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
									throw CodeSegmentOverrun(iIP, aCode[iIP].reserved, szTMP);
								}
								oVMCodeStack.PushAddress(iIP + 1);
								iIP = iNewIP;
							}
							else
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}
						}
						break;

						// Illegal opcode called
						default:
							{
								UINT_32 iDataSize = 0;
								CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);
								throw IllegalOpcode(iIP, iOpCode, aCode[iIP].reserved, szTMP);
							}
					} // End inner switch branch
				} // End default switch branch
			} // End switch
			++iExecutedSteps;
		}
	}
	catch (StackOverflow  & e)
	{
		// Avoid MS VC warning "unised variable"; using paragma is not effective solution
		UINT_32 iTMP = e.GetIP() * 0;

		UINT_32 iDataSize = 0;
		CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);

		throw StackOverflow(iTMP + iIP, aCode[iIP].reserved, szTMP);
	}
	catch (StackUnderflow &e)
	{
		// Avoid MS VC warning "unised variable"; using paragma is not effective solution
		UINT_32 iTMP = e.GetIP() * 0;

		UINT_32 iDataSize = 0;
		CCHAR_P szTMP = oMemoryCore.static_text.GetData(VMDebugInfo(aCode[iIP].reserved).GetDescrId(), iDataSize);

		throw StackUnderflow(iTMP + iIP, aCode[iIP].reserved, szTMP);
	}

return 0;
}

//
// Run one instruction
//
//INT_32 VM::Step(VMInstruction * aCode, const UINT_32 & iCodeLength, UINT_32 & iIP)
//{
//return -1;
//}

//
// Reset virtual machine state
//
INT_32 VM::Reset()
{
	AR = CDT();
	BR = CDT();
	CR = CDT();
	DR = CDT();
	ER = CDT();
	FR = CDT();
	GR = CDT();
	HR = CDT();

	iFlags = 0;

	oVMArgStack.Reset();
	oVMCodeStack.Reset();

return 0;
}

//
// A destructor
//
VM::~VM() throw()
{
	if (aCallTranslationMap != 0) { delete [] aCallTranslationMap; }

}

} // namespace CTPP
// End.
