/*-
 * Copyright (c) 2004 - 2010 CTPP Team
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the CTPP Team nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      VMSTDLibFunctions.cpp
 *
 * $CTPP$
 */

#include "VMSTDLibFunctions.hpp"

#include "CDT.hpp"
#include "CTPP2Util.hpp"
#include "CTPP2Sprintf.hpp"
#include "OutputCollector.hpp"
#include "StaticData.hpp"
#include "StaticText.hpp"
#include "SyscallFactory.hpp"

#include <errno.h>
#include <math.h>

#include <stdio.h>

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include <time.h>

#ifdef GETTEXT_SUPPORT
#include <libintl.h>
#endif // GETTEXT_SUPPORT

#if defined(MD5_SUPPORT) && !defined(WIN32)
    #ifdef MD5_WITHOUT_OPENSSL
        #include <md5.h>
        #define MD5_Init    MD5Init
        #define MD5_Update  MD5Update
        #define MD5_Final   MD5Final
    #else
        #include <openssl/md5.h>
    #endif
#endif // defined(MD5_SUPPORT) && !defined(WIN32)

#ifdef _MSC_VER
	#define random()        rand()
	#define INT_64(x)       (INT_64)(x)
	#define srandomdev()    srand( (unsigned)time(NULL) );
#endif

namespace CTPP // C++ Template Engine
{

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnEmitter
//

//
// Constructor
//
FnEmitter::FnEmitter(): pCDT(NULL)
{
	;;
}

//
// Pre-execution handler setup
//
INT_32 FnEmitter::PreExecuteSetup(OutputCollector          & oCollector,
                                  CDT                      & oCDT,
                                  const ReducedStaticText  & oSyscalls,
                                  const ReducedStaticData  & oStaticData,
                                  const ReducedStaticText  & oStaticText)
{
	pCDT = &oCDT;
return 0;
}

//
// Handler
//
INT_32 FnEmitter::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	oCDTRetVal = *pCDT;
return 0;
}

//
// Get function name
//
CCHAR_P FnEmitter::GetName() { return CTPP2_INT_HANDLER_PREFIX "_emitter"; }

//
// A destructor
//
FnEmitter::~FnEmitter() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnOutput
//

//
// Constructor
//
FnOutput::FnOutput(): pCollector(NULL) { ;; }

//
// Pre-execution handler setup
//
INT_32 FnOutput::PreExecuteSetup(OutputCollector          & oCollector,
                                 CDT                      & oCDT,
                                 const ReducedStaticText  & oSyscalls,
                                 const ReducedStaticData  & oStaticData,
                                 const ReducedStaticText  & oStaticText)
{
	pCollector = &oCollector;
return 0;
}

//
// Handler
//
INT_32 FnOutput::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	for (INT_32 iI = iArgNum - 1; iI >= 0; --iI)
	{
		std::string sTMP(aArguments[iI].GetString());
		if (pCollector -> Collect(sTMP.c_str(), sTMP.length()) == -1) { return -1; }
	}
return 0;
}

//
// Get function name
//
CCHAR_P FnOutput::GetName() { return CTPP2_INT_HANDLER_PREFIX "_output"; }

//
// A destructor
//
FnOutput::~FnOutput() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnOutputText
//

//
// Constructor
//
FnOutputText::FnOutputText(): pCollector(NULL),
                              pStaticText(NULL)
{
	;;
}

//
// Pre-execution handler setup
//
INT_32 FnOutputText::PreExecuteSetup(OutputCollector          & oCollector,
                                     CDT                      & oCDT,
                                     const ReducedStaticText  & oSyscalls,
                                     const ReducedStaticData  & oStaticData,
                                     const ReducedStaticText  & oStaticText)
{
	pCollector  = &oCollector;
	pStaticText = &oStaticText;
return 0;
}

//
// Handler
//
INT_32 FnOutputText::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 1)
	{
		UINT_32 iDatalength = 0;
		CCHAR_P sTMP = pStaticText -> GetData(UINT_32(aArguments[0].GetInt()), iDatalength);
		if (pCollector -> Collect(sTMP, iDatalength) == -1) { return -1; }

		return 0;
	}
return -1;
}

//
// Get function name
//
CCHAR_P FnOutputText::GetName() { return CTPP2_INT_HANDLER_PREFIX "_output_text"; }

//
// A destructor
//
FnOutputText::~FnOutputText() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnOutputInt
//

//
// Constructor
//
FnOutputInt::FnOutputInt(): pCollector(NULL),
                            pStaticData(NULL)
{
	;;
}

//
// Pre-execution handler setup
//
INT_32 FnOutputInt::PreExecuteSetup(OutputCollector          & oCollector,
                                    CDT                      & oCDT,
                                    const ReducedStaticText  & oSyscalls,
                                    const ReducedStaticData  & oStaticData,
                                    const ReducedStaticText  & oStaticText)
{
	pCollector  = &oCollector;
	pStaticData = &oStaticData;
return 0;
}

//
// Handler
//
INT_32 FnOutputInt::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	char sBuffer[2048 + 1];
	if (iArgNum == 1)
	{
		long long iData = pStaticData -> GetInt(UINT_32(aArguments[0].GetInt()));

		INT_32 iChars = snprintf(sBuffer, 2048, "%lli", iData);
		if (iChars < 0) { return -1; }
		if (pCollector -> Collect(sBuffer, iChars) == -1) { return -1; }

		return 0;
	}
return -1;
}

//
// Get function name
//
CCHAR_P FnOutputInt::GetName() { return CTPP2_INT_HANDLER_PREFIX "_output_int"; }

//
// A destructor
//
FnOutputInt::~FnOutputInt() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnOutputFloat
//

//
// Constructor
//
FnOutputFloat::FnOutputFloat(): pCollector(NULL),
                                pStaticData(NULL)
{
	;;
}

//
// Pre-execution handler setup
//
INT_32 FnOutputFloat::PreExecuteSetup(OutputCollector          & oCollector,
                                      CDT                      & oCDT,
                                      const ReducedStaticText  & oSyscalls,
                                      const ReducedStaticData  & oStaticData,
                                      const ReducedStaticText  & oStaticText)
{
	pCollector  = &oCollector;
	pStaticData = &oStaticData;
return 0;
}

//
// Handler
//
INT_32 FnOutputFloat::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	char sBuffer[2048 + 1];
	if (iArgNum == 1)
	{
		double dData = pStaticData -> GetFloat(UINT_32(aArguments[0].GetInt()));

		INT_32 iChars = snprintf(sBuffer, 2048, "%f", dData);
		if (iChars < 0) { return -1; }
		if (pCollector -> Collect(sBuffer, iChars) == -1) { return -1; }

		return 0;
	}
return -1;
}

//
// Get function name
//
CCHAR_P FnOutputFloat::GetName() { return CTPP2_INT_HANDLER_PREFIX "_output_float"; }

//
// A destructor
//
FnOutputFloat::~FnOutputFloat() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnSprintf
//

//
// Constructor
//
FnSprintf::FnSprintf()
{
	;;
}

//
// Handler
//
INT_32 FnSprintf::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 0) { return -1; }

	CDT oArgs(CDT::ARRAY_VAL);
	INT_32 iPos = iArgNum - 2;

	while (iPos >= 0)
	{
		oArgs.PushBack(aArguments[iPos]);
		--iPos;
	}

	std::string sResult;
	FormatString(aArguments[iArgNum - 1].GetString(), sResult, oArgs);
	oCDTRetVal = sResult;
return 0;
}

//
// Get function name
//
CCHAR_P FnSprintf::GetName() { return "sprintf"; }

//
// A destructor
//
FnSprintf::~FnSprintf() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnURIEscape
//

//
// Constructor
//
FnURIEscape::FnURIEscape()
{
	;;
}

//
// Handler
//
INT_32 FnURIEscape::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 1)
	{
		oCDTRetVal = URIEscape(aArguments[0].GetString());
		return 0;
	}

return -1;
}

//
// Get function name
//
CCHAR_P FnURIEscape::GetName() { return "uriescape"; }

//
// A destructor
//
FnURIEscape::~FnURIEscape() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnURLEscape
//

//
// Constructor
//
FnURLEscape::FnURLEscape()
{
	;;
}

//
// Handler
//
INT_32 FnURLEscape::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 1)
	{
		oCDTRetVal = URLEscape(aArguments[0].GetString());
		return 0;
	}

return -1;
}

//
// Get function name
//
CCHAR_P FnURLEscape::GetName() { return "urlescape"; }

//
// A destructor
//
FnURLEscape::~FnURLEscape() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnHTMLEscape
//

//
// Constructor
//
FnHTMLEscape::FnHTMLEscape()
{
	;;
}

//
// Handler
//
INT_32 FnHTMLEscape::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 1) { return -1; }

	oCDTRetVal = HTMLEscape(aArguments[0].GetString());

return 0;
}

//
// Get function name
//
CCHAR_P FnHTMLEscape::GetName() { return "htmlescape"; }

//
// A destructor
//
FnHTMLEscape::~FnHTMLEscape() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnXMLEscape
//

//
// Constructor
//
FnXMLEscape::FnXMLEscape()
{
	;;
}

//
// Handler
//
INT_32 FnXMLEscape::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 1) { return -1; }

	oCDTRetVal = XMLEscape(aArguments[0].GetString());

return 0;
}

//
// Get function name
//
CCHAR_P FnXMLEscape::GetName() { return "xmlescape"; }

//
// A destructor
//
FnXMLEscape::~FnXMLEscape() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnWMLEscape
//

//
// Constructor
//
FnWMLEscape::FnWMLEscape()
{
	;;
}

//
// Handler
//
INT_32 FnWMLEscape::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 1) { return -1; }

	oCDTRetVal = WMLEscape(aArguments[0].GetString());

return 0;
}

//
// Get function name
//
CCHAR_P FnWMLEscape::GetName() { return "wmlescape"; }

//
// A destructor
//
FnWMLEscape::~FnWMLEscape() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnJSEscape
//

//
// Constructor
//
FnJSEscape::FnJSEscape()
{
	;;
}

//
// Handler
//
INT_32 FnJSEscape::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 1) { return -1; }

	const CDT & oCDT = aArguments[0];
	switch (oCDT.GetType())
	{
		case CDT::UNDEF:
			oCDTRetVal = "null";
			break;

		case CDT::INT_VAL:
		case CDT::REAL_VAL:
		case CDT::POINTER_VAL:
			oCDTRetVal = oCDT.GetString();
			break;

		case CDT::STRING_VAL:
			oCDTRetVal.Append(EscapeJSONString(oCDT.GetString(), true));
			break;
		default:
			oCDTRetVal = "";
	}

return 0;
}

//
// Get function name
//
CCHAR_P FnJSEscape::GetName() { return "jsescape"; }

//
// A destructor
//
FnJSEscape::~FnJSEscape() throw() { ;; }


// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnGetText
//
#ifdef GETTEXT_SUPPORT

//
// Constructor
//
FnGetText::FnGetText()
{
	;;
}

//
// Handler
//
INT_32 FnGetText::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 1)
	{
		oCDTRetVal = gettext(aArguments[0].GetString().c_str());
		return 0;
	}
	else if (iArgNum == 2)
	{
		oCDTRetVal = dgettext(aArguments[1].GetString().c_str(), aArguments[0].GetString().c_str());
		return 0;
	}
	else if (iArgNum == 3)
	{
		oCDTRetVal = dcgettext(aArguments[2].GetString().c_str(), aArguments[1].GetString().c_str(), INT_32(aArguments[3].GetInt()));
		return 0;
	}

return -1;
}

//
// Get function name
//
CCHAR_P FnGetText::GetName() { return "gettext"; }

//
// A destructor
//
FnGetText::~FnGetText() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnGetText_
//

//
// Constructor
//
FnGetText_::FnGetText_()
{
	;;
}

//
// Get function name
//
CCHAR_P FnGetText_::GetName() { return "_"; }

//
// A destructor
//
FnGetText_::~FnGetText_() throw() { ;; }

#endif // GETTEXT_SUPPORT

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnInSet
//

//
// Constructor
//
FnInSet::FnInSet()
{
	;;
}

//
// Handler
//
INT_32 FnInSet::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum <= 1) { return -1; }

	INT_32 iI = iArgNum - 1;

	CDT & oCDT1 = aArguments[iI];

	--iI;
	for (; iI >= 0; --iI)
	{
		if (oCDT1 == aArguments[iI]) { oCDTRetVal = 1; return 0; }
	}
	oCDTRetVal = 0;
return 0;
}

//
// Get function name
//
CCHAR_P FnInSet::GetName() { return "in_set"; }

//
// A destructor
//
FnInSet::~FnInSet() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnHrefParam
//

//
// Constructor
//
FnHrefParam::FnHrefParam()
{
	;;
}

//
// Handler
//
INT_32 FnHrefParam::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 2) { return -1; }

	oCDTRetVal = URLEscape(aArguments[1].GetString());
	oCDTRetVal.Append("=");
	oCDTRetVal += URLEscape(aArguments[0].GetString());
	oCDTRetVal.Append("&");

return 0;
}

//
// Get function name
//
CCHAR_P FnHrefParam::GetName() { return "href_param"; }

//
// A destructor
//
FnHrefParam::~FnHrefParam() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnFormParam
//

//
// Constructor
//
FnFormParam::FnFormParam()
{
	;;
}

//
// Handler
//
INT_32 FnFormParam::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 2) { return -1; }

	oCDTRetVal = "<input type=\"hidden\" name=\"";
	oCDTRetVal += HTMLEscape(aArguments[1].GetString());
	oCDTRetVal.Append("\" value=\"");
	oCDTRetVal += HTMLEscape(aArguments[0].GetString());
	oCDTRetVal.Append("\" />");

return 0;
}

//
// Get function name
//
CCHAR_P FnFormParam::GetName() { return "form_param"; }

//
// A destructor
//
FnFormParam::~FnFormParam() throw() { ;; }


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnDateFormat
//

//
// Constructor
//
FnDateFormat::FnDateFormat()
{
	;;
}

//
// Handler
//
INT_32 FnDateFormat::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// Only 2 arguments
	if (iArgNum != 2) { return -1; }
	// Temp
	CHAR_8 szBuffer[CTPP_ESCAPE_BUFFER_LEN + 1];

	CDT & oTMP = aArguments[1];
	time_t iTime = 0;
	if      (oTMP.GetType() <= CDT::REAL_VAL)   { iTime = oTMP.GetInt(); }
	else if (oTMP.GetType() == CDT::STRING_VAL)
	{
		long long iLL = 0;
		sscanf(oTMP.GetString().c_str(), "%lld", &iLL);
		iTime = INT_64(iLL);
	}

	const struct tm * pTM = localtime(&iTime);

	if (strftime(szBuffer, CTPP_ESCAPE_BUFFER_LEN, aArguments[0].GetString().c_str(), pTM) == 0) { return -1; }

	oCDTRetVal = szBuffer;

return 0;
}

//
// Get function name
//
CCHAR_P FnDateFormat::GetName() { return "date_format"; }

//
// A destructor
//
FnDateFormat::~FnDateFormat() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnBase64Encode
//

//
// Constructor
//
FnBase64Encode::FnBase64Encode()
{
	;;
}

//
// Handler
//
INT_32 FnBase64Encode::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// Only 1 argument
	if (iArgNum != 1) { return -1; }

	oCDTRetVal = Base64Encode(aArguments[0].GetString());

return 0;
}

//
// Get function name
//
CCHAR_P FnBase64Encode::GetName() { return "base64_encode"; }

//
// A destructor
//
FnBase64Encode::~FnBase64Encode() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnBase64Decode
//

//
// Constructor
//
FnBase64Decode::FnBase64Decode()
{
	;;
}

//
// Handler
//
INT_32 FnBase64Decode::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// Only 1 argument
	if (iArgNum != 1) { return -1; }

	oCDTRetVal = Base64Decode(aArguments[0].GetString());

return 0;
}

//
// Get function name
//
CCHAR_P FnBase64Decode::GetName() { return "base64_decode"; }

//
// A destructor
//
FnBase64Decode::~FnBase64Decode() throw() { ;; }

#ifdef MD5_SUPPORT

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnHMACMD5
//

//
// Constructor
//
FnHMACMD5::FnHMACMD5()
{
#ifdef WIN32
	// http://msdn.microsoft.com/en-us/library/aa382375(VS.85).aspx
	// Get a handle to the default PROV_RSA_FULL provider.
	if(CryptAcquireContext(&hCryptProv, NULL, NULL, PROV_RSA_FULL, 0))
	{
		// Okay
	}
	else
	{
		if (GetLastError() == NTE_BAD_KEYSET)
		{
			// No default container was found. Attempt to create it.
			if(CryptAcquireContext(&hCryptProv, NULL, NULL, PROV_RSA_FULL, CRYPT_NEWKEYSET))
			{
				// Okay
			}
			else
			{
				throw CTPPUnixException("Could not create the default key container.", GetLastError());
			}
		}
		else
		{
			throw CTPPUnixException("A general error running CryptAcquireContext", GetLastError());
		}
	}
#else
	;;
#endif
}

//
// Handler
//
#ifndef WIN32
INT_32 FnHMACMD5::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	static CHAR_8 aHex[] = "0123456789abcdef";
	CHAR_8 szMD5[32];
	// >0 arguments need
	if (iArgNum != 2) { return -1; }

	MD5_CTX   oContext;
	UCHAR_8  sIPad[65];
	UCHAR_8  sOPad[65];

	// Construct key
	std::string sKey = aArguments[0].GetString();
	// If key is longer than 64 bytes reset it to key=MD5(key)
	if (sKey.size() > 64)
	{
		UCHAR_8  sTempKey[16];
		MD5_Init(&oContext);
		MD5_Update(&oContext, sKey.data(), sKey.size());
		MD5_Final(sTempKey, &oContext);

		sKey.assign((CCHAR_P)sTempKey, 16);
	}

	/*
	 * MD5(K XOR opad, MD5(K XOR ipad, text))
	 *
	 * where K is an n byte key
	 * ipad is the byte 0x36 repeated 64 times
	 * opad is the byte 0x5c repeated 64 times
	 * and text is the data being protected
	 */
	UINT_32 iI = 0;
	for(; iI < sKey.size(); ++iI)
	{
		sIPad[iI] = 0x36 ^ UCHAR_8(sKey[iI]);
		sOPad[iI] = 0x5C ^ UCHAR_8(sKey[iI]);
	}

	for(; iI < 64; ++iI)
	{
		sIPad[iI] = 0x36;
		sOPad[iI] = 0x5C;
	}

	unsigned char sDigest[16];

	std::string sSource = aArguments[1].GetString();
	// Inner MD5
	MD5_Init(&oContext);
	MD5_Update(&oContext, sIPad, 64);
	MD5_Update(&oContext, sSource.data(), sSource.size());
	MD5_Final(sDigest, &oContext);

	// Outer MD5
	MD5_Init(&oContext);
	MD5_Update(&oContext, sOPad, 64);
	MD5_Update(&oContext, sDigest, 16);
	MD5_Final(sDigest, &oContext);

	INT_32 iJ = 0;
	for (iI = 0; iI < 16; ++iI)
	{
		szMD5[ iJ++] = aHex[ (sDigest[iI] >> 4 )& 0x0F ];
		szMD5[ iJ++] = aHex[  sDigest[iI]       & 0x0F ];
	}

	// Okay
	oCDTRetVal = std::string(szMD5, 32);

return 0;
}
#else // WIN32
INT_32 FnHMACMD5::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	static CHAR_8 aHex[] = "0123456789abcdef";
	CHAR_8 szMD5[32];
	// >0 arguments need
	if (iArgNum != 2 || hCryptProv == 0) { return -1; }

	HCRYPTHASH hHash;
	UCHAR_8  sIPad[65];
	UCHAR_8  sOPad[65];

	// Construct key
	std::string sKey = aArguments[0].GetString();
	// If key is longer than 64 bytes reset it to key=MD5(key)
	if (sKey.size() > 64)
	{
		if (::CryptCreateHash(hCryptProv, CALG_MD5, 0, 0, &hHash) == 0) { return -1; }

		if (::CryptHashData(hHash, (BYTE *)sKey.c_str(), (DWORD)sKey.size(), 0) == 0)
		{
			::CryptDestroyHash(hHash);
			return -1;
		}

		BYTE sDigest[16];
		DWORD dwCount = 16;
		if (::CryptGetHashParam(hHash, HP_HASHVAL, sDigest, &dwCount, 0) == 0)
		{
			::CryptDestroyHash(hHash);
			return -1;
		}
		::CryptDestroyHash(hHash);
		sKey.assign((CCHAR_P)sDigest, 16);
	}

	/*
	 * MD5(K XOR opad, MD5(K XOR ipad, text))
	 *
	 * where K is an n byte key
	 * ipad is the byte 0x36 repeated 64 times
	 * opad is the byte 0x5c repeated 64 times
	 * and text is the data being protected
	 */
	UINT_32 iI = 0;
	for(; iI < sKey.size(); ++iI)
	{
		sIPad[iI] = 0x36 ^ UCHAR_8(sKey[iI]);
		sOPad[iI] = 0x5C ^ UCHAR_8(sKey[iI]);
	}

	for(; iI < 64; ++iI)
	{
		sIPad[iI] = 0x36;
		sOPad[iI] = 0x5C;
	}


	BYTE sDigest[16];
	DWORD dwCount = 16;

	std::string sSource = aArguments[1].GetString();
	// Inner MD5
	if (::CryptCreateHash(hCryptProv, CALG_MD5, 0, 0, &hHash) == 0) { return -1; }

	if (::CryptHashData(hHash, (BYTE *)sIPad, (DWORD)64, 0) == 0)
	{
		::CryptDestroyHash(hHash);
		return -1;
	}

	if (::CryptHashData(hHash, (BYTE *)sSource.data(), (DWORD)sSource.size(), 0) == 0)
	{
		::CryptDestroyHash(hHash);
		return -1;
	}

	if (::CryptGetHashParam(hHash, HP_HASHVAL, sDigest, &dwCount, 0) == 0)
	{
		::CryptDestroyHash(hHash);
		return -1;
	}
	::CryptDestroyHash(hHash);

	if (::CryptCreateHash(hCryptProv, CALG_MD5, 0, 0, &hHash) == 0) { return -1; }

	if (::CryptHashData(hHash, (BYTE *)sOPad, (DWORD)64, 0) == 0)
	{
		::CryptDestroyHash(hHash);
		return -1;
	}

	if (::CryptHashData(hHash, (BYTE *)sDigest, (DWORD)16, 0) == 0)
	{
		::CryptDestroyHash(hHash);
		return -1;
	}

	if (::CryptGetHashParam(hHash, HP_HASHVAL, sDigest, &dwCount, 0) == 0)
	{
		::CryptDestroyHash(hHash);
		return -1;
	}
	::CryptDestroyHash(hHash);

	INT_32 iJ = 0;
	for (iI = 0; iI < 16; ++iI)
	{
		szMD5[ iJ++] = aHex[ (sDigest[iI] >> 4 )& 0x0F ];
		szMD5[ iJ++] = aHex[  sDigest[iI]       & 0x0F ];
	}

	// Okay
	oCDTRetVal = std::string(szMD5, 32);

return 0;
}
#endif // WIN32
//
// Get function name
//
CCHAR_P FnHMACMD5::GetName() { return "hmac_md5"; }

//
// A destructor
//
FnHMACMD5::~FnHMACMD5() throw()
{
#ifdef WIN32
	if (hCryptProv != 0) { ::CryptReleaseContext(hCryptProv, 0); }
#else
	;;
#endif
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnMD5
//

//
// Constructor
//
FnMD5::FnMD5()
{
#ifdef WIN32
	// http://msdn.microsoft.com/en-us/library/aa382375(VS.85).aspx
	// Get a handle to the default PROV_RSA_FULL provider.
	if(CryptAcquireContext(&hCryptProv, NULL, NULL, PROV_RSA_FULL, 0))
	{
		// Okay
	}
	else
	{
		if (GetLastError() == NTE_BAD_KEYSET)
		{
			// No default container was found. Attempt to create it.
			if(CryptAcquireContext(&hCryptProv, NULL, NULL, PROV_RSA_FULL, CRYPT_NEWKEYSET))
			{
				// Okay
			}
			else
			{
				throw CTPPUnixException("Could not create the default key container.", GetLastError());
			}
		}
		else
		{
			throw CTPPUnixException("A general error running CryptAcquireContext", GetLastError());
		}
	}
#else
	;;
#endif
}

//
// Handler
//
#ifndef WIN32
INT_32 FnMD5::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	static CHAR_8 aHex[] = "0123456789abcdef";
	CHAR_8 szMD5[32];
	// >0 arguments need
	if (iArgNum == 0) { return -1; }

	MD5_CTX oMD5Context;
	MD5_Init(&oMD5Context);

	INT_32 iI = iArgNum - 1;
	for (; iI >= 0; --iI)
	{
		// Data
		const std::string & sTMP = aArguments[iI].GetString();
		CCHAR_P szData = sTMP.c_str();
		// Data length
		const UINT_32 iDataLength = sTMP.size();

		MD5_Update(&oMD5Context, (const unsigned char *)szData, iDataLength);
	}

	unsigned char sDigest[16];
	MD5_Final(sDigest, &oMD5Context);

	INT_32 iJ = 0;
	for (iI = 0; iI < 16; ++iI)
	{
		szMD5[ iJ++] = aHex[ (sDigest[iI] >> 4 )& 0x0F ];
		szMD5[ iJ++] = aHex[  sDigest[iI]       & 0x0F ];
	}

	// Okay
	oCDTRetVal = std::string(szMD5, 32);

return 0;

}
#else // WIN32

//
// Handler, using Win32 CryptoAPI
//
INT_32 FnMD5::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	static CHAR_8 aHex[] = "0123456789abcdef";
	CHAR_8 szMD5[32];
	HCRYPTHASH hHash;

	// >0 arguments need
	if (iArgNum == 0 || hCryptProv == 0) { fprintf(stderr, "FnMD5 iArgNum = %d hCryptProv = %d\n", iArgNum, hCryptProv); return -1; }

	if (::CryptCreateHash(hCryptProv, CALG_MD5, 0, 0, &hHash) == 0) { fprintf(stderr, "CryptCreateHash\n"); return -1; }

	INT_32 iI = iArgNum - 1;
	for (; iI >= 0; --iI)
	{
		// Data
		const std::string & sTMP = aArguments[iI].GetString();
		if (::CryptHashData(hHash, (BYTE *)sTMP.c_str(), (DWORD)sTMP.size(), 0) == 0)
		{
			fprintf(stderr, "CryptHashData\n");
			::CryptDestroyHash(hHash);
			return -1;
		}
	}

	BYTE sDigest[16];
	DWORD dwCount = 16;
	if (::CryptGetHashParam(hHash, HP_HASHVAL, sDigest, &dwCount, 0) == 0)
	{
		fprintf(stderr, "CryptGetHashParam\n");
		::CryptDestroyHash(hHash);
		return -1;
	}

	INT_32 iJ = 0;
	for (iI = 0; iI < 16; ++iI)
	{
		szMD5[ iJ++] = aHex[ (sDigest[iI] >> 4 )& 0x0F ];
		szMD5[ iJ++] = aHex[  sDigest[iI]       & 0x0F ];
	}

	::CryptDestroyHash(hHash);

	// Okay
	oCDTRetVal = std::string(szMD5, 32);

return 0;
}
#endif // WIN32

//
// Get function name
//
CCHAR_P FnMD5::GetName() { return "md5"; }

//
// A destructor
//
FnMD5::~FnMD5() throw()
{
#ifdef WIN32
	if (hCryptProv != 0) { ::CryptReleaseContext(hCryptProv, 0); }
#else
	;;
#endif
}
#endif // MD5_SUPPORT

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnIconv
//
#ifdef ICONV_SUPPORT

#ifndef C_ICONV_DISCARD_ILSEQ
#define C_ICONV_DISCARD_ILSEQ    0x00000001
#endif

#ifndef C_ICONV_TRANSLITERATE
#define C_ICONV_TRANSLITERATE    0x00000002
#endif

//
// Constructor
//
FnIconv::FnIconv()
{
	;;
}

//
// Handler
//
INT_32 FnIconv::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	UINT_32 iMyArgNum  = iArgNum;

	// 3 or 4 arguments need
	if (iMyArgNum != 3 && iMyArgNum != 4) { return -1; }

	// Arg 3: flags
	// Arg 2: destination charset
	// Arg 1: source charset
	// Arg 0: string to convert

	UINT_32 iFlags = 0;

#ifdef ICONV_DISCARD_ILSEQ
	iFlags |= C_ICONV_DISCARD_ILSEQ;
#endif

#ifdef ICONV_TRANSLITERATE
	iFlags |= C_ICONV_TRANSLITERATE;
#endif

	if (iMyArgNum == 4)
	{
		const std::string & sFlags = aArguments[0].GetString();
		for (UINT_32 iPos = 0; iPos < sFlags.size(); ++iPos)
		{
			switch (sFlags[iPos])
			{
				// Discard illegal sequence and continue
				case 'i':
				case 'I':
					iFlags |= C_ICONV_DISCARD_ILSEQ;
					break;

				// Enable transliteration
				case 't':
				case 'T':
					iFlags |= C_ICONV_TRANSLITERATE;
					break;

				default:
					return -1;
			}
		}
	}

	const std::string & sTo   = aArguments[--iMyArgNum].GetString();
	const std::string & sFrom = aArguments[--iMyArgNum].GetString();
	const std::string & sWhat = aArguments[--iMyArgNum].GetString();

	std::string sFromTo(sFrom);
	sFromTo.append(sTo);

	iconv_t oIconvConverter = (iconv_t)(-1);

	std::map<std::string, iconv_t>::iterator itmIconvMap = mIconvMap.find(sFromTo);
	if (itmIconvMap != mIconvMap.end()) { oIconvConverter = itmIconvMap -> second; }
	// Try to open iconv converter
	else
	{
		oIconvConverter = iconv_open(sFrom.c_str(), sTo.c_str());

		if (oIconvConverter != (iconv_t)(-1)) { mIconvMap[sFromTo] = oIconvConverter; }
		else                                  { return -1;                            }
	}

#if (_LIBICONV_VERSION >= 0x0108)
	int iFlag = 1;
	// Discard illegal characters
	if (iFlags & C_ICONV_DISCARD_ILSEQ)
	{
		if (iconvctl(oIconvConverter, ICONV_SET_DISCARD_ILSEQ, &iFlag) == -1) { return -1; }
	}

	// nable transliteration in the conver-sion
	if (iFlags & C_ICONV_TRANSLITERATE)
	{
		if (iconvctl(oIconvConverter, ICONV_SET_TRANSLITERATE, &iFlag) == -1) { return -1; }
	}
#endif

	// Allocate memory
	size_t iSrcLength     = sWhat.size();
	size_t iDstLength     = iSrcLength * 2;
	for (;;)
	{

		char       * aDstData = new char[iDstLength];
#if defined(linux) || defined(__APPLE__)
		char       * aSrcData = (char *)sWhat.c_str();
#else
		const char * aSrcData = sWhat.c_str();
#endif
		char * aDstTMP       = aDstData;
		size_t iDstLengthTMP = iDstLength;
		size_t iResult       = iconv(oIconvConverter, &aSrcData, &iSrcLength, &aDstTMP, &iDstLengthTMP);

		if (iResult == (size_t)-1)
		{
			iconv(oIconvConverter, NULL, NULL, NULL, NULL);
			delete [] aDstData;
			if (errno != E2BIG) { return -1; }

			iDstLength *= 2;
			iSrcLength  = sWhat.size();
		}
		else
		{
			// Okay
			oCDTRetVal = std::string(aDstData, aDstTMP - aDstData);
			delete [] aDstData;
			break;
		}
	}

return 0;

}

//
// Get function name
//
CCHAR_P FnIconv::GetName() { return "iconv"; }

//
// A destructor
//
FnIconv::~FnIconv() throw()
{
	std::map<std::string, iconv_t>::iterator itmIconvMap = mIconvMap.begin();
	while (itmIconvMap != mIconvMap.end())
	{
		iconv_close(itmIconvMap -> second);
		++itmIconvMap;
	}
}

#endif // ICONV_SUPPORT

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnCast
//

//
// Constructor
//
FnCast::FnCast()
{
	;;
}

//
// Handler
//
INT_32 FnCast::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// 2 arguments need
	if (iArgNum != 2) { return -1; }

	// Arg 1: destination type
	// Arg 0: source type
	const std::string & sWhat = aArguments[1].GetString();
	if (sWhat.length() == 0) { return -1; }

	CDT & oTMP = aArguments[0];

	switch(sWhat[0])
	{
		// S[tring]
		case 'S':
		case 's':
			oCDTRetVal = oTMP.GetString();
			return 0;

		// O[ct[al]]
		case 'o':
		case 'O':
			if      (oTMP.GetType() <= CDT::REAL_VAL)   { oCDTRetVal = oTMP.GetInt(); return 0; }
			else if (oTMP.GetType() == CDT::STRING_VAL)
			{
				unsigned long long iLL = 0;
				sscanf(oTMP.GetString().c_str(), "%llo", &iLL);
				oCDTRetVal = INT_64(iLL);
				return 0;
			}
			return -1;

		// D[ec[imal]]
		case 'd':
		case 'D':
			if      (oTMP.GetType() <= CDT::REAL_VAL)   { oCDTRetVal = oTMP.GetInt(); return 0; }
			else if (oTMP.GetType() == CDT::STRING_VAL)
			{
				long long iLL = 0;
				sscanf(oTMP.GetString().c_str(), "%lld", &iLL);
				oCDTRetVal = INT_64(iLL);
				return 0;
			}
			return -1;

		// I[nt[eger]]
		case 'i':
		case 'I':
			if      (oTMP.GetType() <= CDT::REAL_VAL)   { oCDTRetVal = oTMP.GetInt(); return 0; }
			else if (oTMP.GetType() == CDT::STRING_VAL)
			{
				long long iLL = 0;
				sscanf(oTMP.GetString().c_str(), "%lli", &iLL);
				oCDTRetVal = INT_64(iLL);
				return 0;
			}
			return -1;

		// H[ex[adecimal]]
		case 'H':
		case 'h':
			if      (oTMP.GetType() <= CDT::REAL_VAL)   { oCDTRetVal = oTMP.GetInt(); return 0; }
			else if (oTMP.GetType() == CDT::STRING_VAL)
			{
				unsigned long long iLL = 0;
				sscanf(oTMP.GetString().c_str(), "%llx", &iLL);
				oCDTRetVal = INT_64(iLL);
				return 0;
			}
			return -1;

		// F[loat]
		case 'F':
		case 'f':
			if      (oTMP.GetType() <= CDT::REAL_VAL)   { oCDTRetVal = oTMP.GetFloat(); return 0; }
			else if (oTMP.GetType() == CDT::STRING_VAL)
			{
				double dVal = 0;
				sscanf(oTMP.GetString().c_str(), "%lg", &dVal);
				oCDTRetVal = dVal;
				return 0;
			}
			return -1;

		// Error
		default:
			return -1;
	}
}

//
// Get function name
//
CCHAR_P FnCast::GetName() { return "cast"; }

//
// A destructor
//
FnCast::~FnCast() throw()
{
	;;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnDefined
//

//
// Constructor
//
FnDefined::FnDefined()
{
	;;
}

//
// Handler
//
INT_32 FnDefined::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// At least one argument need
	if (iArgNum == 0) { return -1; }

	oCDTRetVal = 1;

	INT_32 iI = iArgNum - 1;
	for (; iI >= 0; --iI)
	{
		if (aArguments[iI].GetType() == CDT::UNDEF)
		{
			oCDTRetVal = 0;
			break;
		}
	}

return 0;
}

//
// Get function name
//
CCHAR_P FnDefined::GetName() { return "defined"; }

//
// A destructor
//
FnDefined::~FnDefined() throw()
{
	;;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnSize
//

//
// Constructor
//
FnSize::FnSize()
{
	;;
}

//
// Handler
//
INT_32 FnSize::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// At least one argument need
	if (iArgNum != 1) { return -1; }

	oCDTRetVal = aArguments[0].Size();

return 0;
}

//
// Get function name
//
CCHAR_P FnSize::GetName() { return "size"; }

//
// A destructor
//
FnSize::~FnSize() throw()
{
	;;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnMBSize
//

//
// Constructor
//
FnMBSize::FnMBSize()
{
	;;
}

//
// Handler
//
INT_32 FnMBSize::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// At least one argument need
	if (iArgNum != 1) { return -1; }

	if (aArguments[0].GetType() != CDT::STRING_VAL) { oCDTRetVal = aArguments[0].Size(); }
	else
	{
		const std::string sData = aArguments[0].GetString();

		CCHAR_P  szStart  = sData.data();
		CCHAR_P  szEnd    = szStart + sData.size();
		INT_32   iPos     = 0;
		UINT_32  iCharPos = 0;
		for(;;)
		{
			INT_32 iCharLen = utf_charlen(szStart + iPos, szEnd);
			if (iCharLen == -3) { break; }

			// Check character length
			if (iCharLen < 0) { iCharLen = 1; }

			++iCharPos;
			iPos += iCharLen;
		}
		oCDTRetVal = iCharPos;
	}

return 0;
}

//
// Get function name
//
CCHAR_P FnMBSize::GetName() { return "mb_size"; }

//
// A destructor
//
FnMBSize::~FnMBSize() throw()
{
	;;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnNumFormat
//

//
// Constructor
//
FnNumFormat::FnNumFormat()
{
	;;
}

//
// Handler
//
INT_32 FnNumFormat::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// At least one argument need
	if (iArgNum != 2) { return -1; }

	std::string sTMP = aArguments[0].GetString();

	switch (aArguments[1].GetType())
	{
		// Integer value
		case CDT::INT_VAL:
			{
				CHAR_8 szRes[C_MAX_SPRINTF_LENGTH + 1];
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];

				szRes[C_MAX_SPRINTF_LENGTH] = '\0';

				INT_64 iIntData = aArguments[1].GetInt();

				INT_32 iNegative = 0;
				if (iIntData < 0) { iIntData = -iIntData; iNegative = 1; }

				INT_32 iCharacters = snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "%lld", (long long)iIntData);
				INT_32 iResPos = C_MAX_SPRINTF_LENGTH - 1;

				INT_32 iPos      = 0;
				while (iCharacters >= 0)
				{
					szRes[iResPos] = szBuf[iCharacters];
					if (iPos++ % 3 == 0 && iPos != 1) { szRes[--iResPos] = sTMP[0]; }
					--iCharacters;
					--iResPos;
				}

				++iResPos;
				if (iPos % 3 == 1) { ++iResPos; }

				if (iNegative == 1) { szRes[--iResPos] = '-'; }

				oCDTRetVal = &szRes[iResPos];
			}
			break;

		// Floating point value
		case CDT::REAL_VAL:
			{
				CHAR_8 szRes[C_MAX_SPRINTF_LENGTH + 1];
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];

				szRes[C_MAX_SPRINTF_LENGTH] = '\0';

				W_FLOAT dFloatData = aArguments[1].GetFloat();

				INT_32 iNegative = 0;
				if (dFloatData < 0) { dFloatData = -dFloatData; iNegative = 1; }

				INT_32 iCharacters = snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "%.*G", CTPP_FLOAT_PRECISION, dFloatData);
				INT_32 iResPos = C_MAX_SPRINTF_LENGTH - 1;

				INT_32 iTMP = iCharacters;
				iCharacters = 0;
				while (iCharacters < iTMP && szBuf[iCharacters++] != '.') { ;; }

				while (iTMP >= iCharacters)
				{
					szRes[iResPos] = szBuf[iTMP];
					--iTMP;
					--iResPos;
				}

				if (szBuf[iTMP] == '.') { iCharacters = iTMP; }

				INT_32 iPos = 0;
				while (iCharacters >= 0)
				{
					szRes[iResPos] = szBuf[iCharacters];
					if (iPos++ % 3 == 0 && iPos != 1) { szRes[--iResPos] = sTMP[0]; }
					--iCharacters;
					--iResPos;
				}

				++iResPos;
				if (iPos % 3 == 1) { ++iResPos; }

				if (iNegative == 1) { szRes[--iResPos] = '-'; }

				oCDTRetVal = &szRes[iResPos];
			}
			break;

		// All other types
		default:
			oCDTRetVal = aArguments[1].GetString();
	}

return 0;
}

//
// Get function name
//
CCHAR_P FnNumFormat::GetName() { return "num_format"; }

//
// A destructor
//
FnNumFormat::~FnNumFormat() throw()
{
	;;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnNumeral
//

//
// Constructor
//
FnNumeral::FnNumeral()
{
	;;
}

//
// Handler
//
INT_32 FnNumeral::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// At least two arguments need
	if (iArgNum < 2) { return -1; }

	// Number
	INT_64 iNumber = aArguments[iArgNum - 1].GetInt();

	const std::string sLang = aArguments[iArgNum - 2].GetString();
	if (sLang == "ru")
	{
		if (iArgNum != 5)
		{
			oCDTRetVal = "For Russian language need 5 arguments: NUMERAL(number, language, first, second-fourth, fifth-20th)";
			return 0;
		}

		INT_32 iN = iNumber % 100;
		if ((iN >= 5 && iN <= 20) || iN == 0)
		{
			oCDTRetVal = aArguments[0];
			return 0;
		}

		iN %= 10;
		switch(iN)
		{
			case 0:
				oCDTRetVal = aArguments[0];
				break;
			case 1:
				oCDTRetVal = aArguments[2];
				break;
			case 2:
			case 3:
			case 4:
				oCDTRetVal = aArguments[1];
				break;
			default:
				oCDTRetVal = aArguments[0];
		}
	}
	else if (sLang == "en")
	{
		if (iArgNum != 4)
		{
			oCDTRetVal = "For English language need 4 arguments: NUMERAL(number, language, first, other)";
			return 0;
		}

		if (iNumber == 1) { oCDTRetVal = aArguments[1]; }
		else              { oCDTRetVal = aArguments[0]; }
	}
	else
	{
		oCDTRetVal = std::string("Language `") + sLang + "` unsupported yet";
	}


return 0;
}

//
// Get function name
//
CCHAR_P FnNumeral::GetName() { return "numeral"; }

//
// A destructor
//
FnNumeral::~FnNumeral() throw()
{
	;;
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnObjDump
//

//
// Constructor
//
FnObjDump::FnObjDump(): pCDT(NULL)
{
	;;
}

//
// Pre-execution handler setup
//
INT_32 FnObjDump::PreExecuteSetup(OutputCollector          & oCollector,
                                  CDT                      & oCDT,
                                  const ReducedStaticText  & oSyscalls,
                                  const ReducedStaticData  & oStaticData,
                                  const ReducedStaticText  & oStaticText)
{
	pCDT = &oCDT;
return 0;
}

//
// Handler
//
INT_32 FnObjDump::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 0)
	{
		oCDTRetVal = "Global data object:\n";
		oCDTRetVal.Append(pCDT -> RecursiveDump());
	}
	else
	{
		oCDTRetVal = "";
		// Temp
		CHAR_8 szBuffer[CTPP_ESCAPE_BUFFER_LEN + 1];

		INT_32 iI = iArgNum - 1;
		for (; iI >= 0; --iI)
		{
			snprintf(szBuffer, CTPP_ESCAPE_BUFFER_LEN, "Argument %d:\n", iArgNum - iI);
			oCDTRetVal.Append(szBuffer);
			oCDTRetVal.Append(aArguments[iI].RecursiveDump());
		}
	}
return 0;
}

//
// Get function name
//
CCHAR_P FnObjDump::GetName() { return "obj_dump"; }

//
// A destructor
//
FnObjDump::~FnObjDump() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnVersion
//

//
// Constructor
//
FnVersion::FnVersion() { ;; }

//
// Handler
//
INT_32 FnVersion::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 1 && strcasecmp("full", aArguments[0].GetString().c_str()) == 0)
	{
		oCDTRetVal = "Engine: CTPP2 engine v" CTPP_VERSION " (" CTPP_IDENT "), copyright (c) 2004 - 2010 CTPP Dev. Team;\n"
		             "RuntimeLibrary: CTPP Standard Library v" CTPP_VERSION " (" CTPP_IDENT "), copyright (c) 2007 - 2010 CTPP Dev. Team;\n"
		             "License: BSD-like, see " CTPP_MASTER_SITE_URL ";";
	}
	else
	{
		oCDTRetVal = "CTPP2 engine v" CTPP_VERSION " (" CTPP_IDENT "), copyright (c) 2004 - 2010 CTPP Dev. Team";
	}

return 0;
}

//
// Get function name
//
CCHAR_P FnVersion::GetName() { return "version"; }

//
// A destructor
//
FnVersion::~FnVersion() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnError
//

//
// Constructor
//
FnError::FnError()
{
	;;
}

//
// Handler
//
INT_32 FnError::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal) { return -1; }

//
// Get function name
//
CCHAR_P FnError::GetName() { return "error"; }

//
// A destructor
//
FnError::~FnError() throw()
{
	;;
}

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnBBCode
//

//
// Constructor
//
FnBBCode::FnBBCode() { ;; }

//
// Pre-execution handler setup
//
INT_32 FnBBCode::PreExecuteSetup(OutputCollector          & oCollector,
                                 CDT                      & oCDT,
                                 const ReducedStaticText  & oSyscalls,
                                 const ReducedStaticData  & oStaticData,
                                 const ReducedStaticText  & oStaticText)
{
	//
return 0;
}

//
// Handler
//
INT_32 FnBBCode::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 1) { return -1; }

	std::string sTMP(aArguments[0].GetString());

	//
	// [b]
	// [i]
	// [u]
	// [s]
	// [quote name="username" post="123"]
	// [size]
	// [code]
	// [color]
	// [img]
	// [url]
	// [youtube]
	// [rutube]
	// [right]
	// [left]
	// [center]
	// [spoiler]
	// [topic]
	// [post]
	// [forum]
	//

return 0;
}

//
// Get function name
//
CCHAR_P FnBBCode::GetName() { return "bb_code"; }

//
// A destructor
//
FnBBCode::~FnBBCode() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnDefault
//

//
// Constructor
//
FnDefault::FnDefault()
{
	;;
}

//
// Handler
//
INT_32 FnDefault::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// At least one argument need
	if (iArgNum != 2) { return -1; }

	// Get first arument
	oCDTRetVal = aArguments[1];

	// Set default value from second argument
	if (aArguments[1].GetType() == CDT::UNDEF) { oCDTRetVal = aArguments[0]; }
	else                                       { oCDTRetVal = aArguments[1]; }

return 0;
}

//
// Get function name
//
CCHAR_P FnDefault::GetName() { return "default"; }

//
// A destructor
//
FnDefault::~FnDefault() throw()
{
	;;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnJSON
//

//
// Constructor
//
FnJSON::FnJSON()
{
	;;
}

//
// Handler
//
INT_32 FnJSON::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 1) { return -1; }

	std::string sData;

	CDT2JSON(aArguments[0], sData);

	oCDTRetVal = sData;

return 0;
}

//
// Get function name
//
CCHAR_P FnJSON::GetName() { return "json"; }

//
// A destructor
//
FnJSON::~FnJSON() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnRandom
//

//
// Constructor
//
FnRandom::FnRandom()
{
#if defined(__FreeBSD__) || defined(_MSC_VER)
	srandomdev();
#else
	srandom(time(NULL));
#endif
}

//
// Handler
//
INT_32 FnRandom::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// 0 .. MAX(unsigned long)
	if (iArgNum == 0)
	{
		oCDTRetVal = UINT_64(random());
		return 0;
	}
	// RAND(x) -> 0 .. RAND(x)
	else if (iArgNum == 1)
	{
		const CDT & oFirstArg = aArguments[0];
		// Integer value
		if (oFirstArg.GetType() == CDT::INT_VAL)
		{
			oCDTRetVal = INT_64(random() % oFirstArg.GetInt());
			return 0;
		}
		// Floating point value
		else if (oFirstArg.GetType() == CDT::REAL_VAL)
		{
			oCDTRetVal = W_FLOAT(oFirstArg.GetFloat() * (random() % UINT_64(RAND_MAX)) / UINT_64(RAND_MAX));
			return 0;
		}

		// Invalid data type, just return 0
		oCDTRetVal = 0;
		return 0;
	}
	// RAND(x, y) -> x .. y
	else if (iArgNum == 2)
	{
		const CDT & oFirstArg  = aArguments[1];
		const CDT & oSecondArg = aArguments[0];

		// x and y is integer values
		if (oFirstArg.GetType()  == CDT::INT_VAL &&
		    oSecondArg.GetType() == CDT::INT_VAL)
		{
			const UINT_64 iFirst    = oFirstArg.GetInt();
			const UINT_64 iInterval = oSecondArg.GetInt() - iFirst;

			oCDTRetVal = INT_64(random() % iInterval + iFirst);
			return 0;
		}
		// x or y (or both) are floating point value
		else if ((oFirstArg.GetType() == CDT::REAL_VAL && oSecondArg.GetType() == CDT::INT_VAL)  ||
		         (oFirstArg.GetType() == CDT::INT_VAL  && oSecondArg.GetType() == CDT::REAL_VAL) ||
		         (oFirstArg.GetType() == CDT::REAL_VAL && oSecondArg.GetType() == CDT::REAL_VAL))
		{
			const W_FLOAT dFirst    = oFirstArg.GetFloat();
			const W_FLOAT dInterval = oSecondArg.GetFloat() - dFirst;

			oCDTRetVal = W_FLOAT(dInterval * (random() % UINT_64(RAND_MAX)) / UINT_64(RAND_MAX) + dFirst);
			return 0;
		}

		// Invalid data type, just return 0
		oCDTRetVal = 0;
		return 0;
	}

return -1;
}

//
// Get function name
//
CCHAR_P FnRandom::GetName() { return "random"; }

//
// A destructor
//
FnRandom::~FnRandom() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnLog
//

//
// Constructor
//
FnLog::FnLog()
{
	;;
}

//
// Handler
//
INT_32 FnLog::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// Natural logarithm
	if (iArgNum == 1)
	{
		// NaN
		oCDTRetVal = log(aArguments[0].GetFloat());
		return 0;
	}
	// Logarithm with specified base
	else if (iArgNum == 2)
	{
		const W_FLOAT dBase  = aArguments[0].GetFloat();
		const W_FLOAT dValue = aArguments[1].GetFloat();

		// NaN
		if (dBase <= 0) { return (INT_32)log((W_FLOAT)-1); }

		oCDTRetVal = log(dValue) / log(dBase);
		return 0;
	}

return -1;
}

//
// Get function name
//
CCHAR_P FnLog::GetName() { return "log"; }

//
// A destructor
//
FnLog::~FnLog() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnMin
//

//
// Constructor
//
FnMin::FnMin()
{
	;;
}

//
// Handler
//
INT_32 FnMin::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 0) { return -1; }

	INT_32  iMinPos = 0;
	W_FLOAT dMinVal = aArguments[0].GetFloat();
	for (UINT_32 iPos = 0; iPos < iArgNum; ++iPos)
	{
		const W_FLOAT dCurVal = aArguments[iPos].GetFloat();
		if (dCurVal < dMinVal) { iMinPos = iPos; dMinVal = dCurVal; }
	}

	oCDTRetVal = aArguments[iMinPos];

return 0;
}

//
// Get function name
//
CCHAR_P FnMin::GetName() { return "min"; }

//
// A destructor
//
FnMin::~FnMin() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnMax
//

//
// Constructor
//
FnMax::FnMax()
{
	;;
}

//
// Handler
//
INT_32 FnMax::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 0) { return -1; }

	INT_32  iMaxPos = 0;
	W_FLOAT dMaxVal = aArguments[0].GetFloat();
	for (UINT_32 iPos = 0; iPos < iArgNum; ++iPos)
	{
		const W_FLOAT dCurVal = aArguments[iPos].GetFloat();
		if (dCurVal > dMaxVal) { iMaxPos = iPos; dMaxVal = dCurVal; }
	}

	oCDTRetVal = aArguments[iMaxPos];

return 0;
}

//
// Get function name
//
CCHAR_P FnMax::GetName() { return "max"; }

//
// A destructor
//
FnMax::~FnMax() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnAvg
//

//
// Constructor
//
FnAvg::FnAvg()
{
	;;
}

//
// Handler
//
INT_32 FnAvg::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum < 2) { return -1; }

	// First Arg 1: destination type
	const std::string & sWhat = aArguments[iArgNum - 1].GetString();
	if (sWhat.length() == 0) { return -1; }

	// http://en.wikipedia.org/wiki/Average
	switch (sWhat[0])
	{
		// Arithmetic
		case 'A':
		case 'a':
			{
				W_FLOAT  dSum = 0.0;
				for(INT_32 iPos = iArgNum - 2; iPos >= 0 ; --iPos)
				{
					dSum += aArguments[iPos].GetFloat();
				}
				oCDTRetVal = dSum / (iArgNum - 1);
			}
			break;
		// Geometric
		case 'G':
		case 'g':
			{
				W_FLOAT  dSum = 1.0;
				for(INT_32 iPos = iArgNum - 2; iPos >= 0 ; --iPos)
				{
					dSum *= aArguments[iPos].GetFloat();
				}
				oCDTRetVal = pow(dSum, 1.0 / (iArgNum - 1));
			}
			break;

		// Harmonic
		case 'H':
		case 'h':
			{
				W_FLOAT  dSum = 0.0;
				for(INT_32 iPos = iArgNum - 2; iPos >= 0 ; --iPos)
				{
					const W_FLOAT dTMP = aArguments[iPos].GetFloat();
					if (dTMP == 0.0) { dSum = 0; break; }
					dSum += 1.0 / dTMP;
				}
				oCDTRetVal = 1.0 * (iArgNum - 1) / dSum;
			}
			break;

		// Quadratic
		case 'Q':
		case 'q':
			{
				W_FLOAT  dSum = 0.0;
				for(INT_32 iPos = iArgNum - 1; iPos >= 0 ; --iPos)
				{
					const W_FLOAT dTMP = aArguments[iPos].GetFloat();
					dSum += dTMP * dTMP;
				}
				oCDTRetVal = sqrt(dSum / (iArgNum - 1));
			}
			break;
		default:
			return -1;
	}

return 0;
}

//
// Get function name
//
CCHAR_P FnAvg::GetName() { return "avg"; }

//
// A destructor
//
FnAvg::~FnAvg() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnTruncate
//

//
// Constructor
//
FnTruncate::FnTruncate()
{
	;;
}

//
// Handler
//
INT_32 FnTruncate::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 2)
	{
		const UINT_32       iMaxLen = UINT_32(aArguments[0].GetInt());
		const std::string   sData   = aArguments[1].GetString();

		if (sData.size() > iMaxLen) { oCDTRetVal = std::string(sData, 0, iMaxLen); }
		else                        { oCDTRetVal = sData;                          }

		return 0;
	}
	else if (iArgNum == 3)
	{
		const UINT_32  iMaxLen = UINT_32(aArguments[1].GetInt());
		std::string    sData   = aArguments[2].GetString();

		if (sData.size() > iMaxLen)
		{
			sData = std::string(sData, 0, iMaxLen);
			sData.append(aArguments[0].GetString());
		}

		oCDTRetVal = sData;
		return 0;
	}

return -1;
}

//
// Get function name
//
CCHAR_P FnTruncate::GetName() { return "truncate"; }

//
// A destructor
//
FnTruncate::~FnTruncate() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnMBTruncate
//

//
// Constructor
//
FnMBTruncate::FnMBTruncate()
{
	;;
}

//
// Handler
//
INT_32 FnMBTruncate::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 2)
	{
		const UINT_32       iMaxLen = UINT_32(aArguments[0].GetInt());
		const std::string   sData   = aArguments[1].GetString();

		CCHAR_P  szStart  = sData.data();
		CCHAR_P  szEnd    = sData.data() + sData.size();
		INT_32   iPos     = 0;
		UINT_32  iCharPos = 0;
		for(;;)
		{
			INT_32 iCharLen = utf_charlen(szStart + iPos, szEnd);
			if (iCharLen == -3) { break; }

			// Check character length
			if (iCharLen < 0) { iCharLen = 1; }
			// Skip errors
			else              { ++iCharPos;   }
			iPos += iCharLen;

			if (iCharPos >= iMaxLen) { break; }
		}

		if (iCharPos == iMaxLen) { oCDTRetVal = std::string(sData, 0, iPos); }
		else                     { oCDTRetVal = sData;                       }

		return 0;
	}
	else if (iArgNum == 3)
	{
		const UINT_32  iMaxLen = UINT_32(aArguments[1].GetInt());
		std::string    sData   = aArguments[2].GetString();

		CCHAR_P  szStart  = sData.data();
		CCHAR_P  szEnd    = sData.data() + sData.size();
		INT_32   iPos     = 0;
		UINT_32  iCharPos = 0;
		for(;;)
		{
			INT_32 iCharLen = utf_charlen(szStart + iPos, szEnd);

			if (iCharLen == -3) { break; }

			// Check character length
			if (iCharLen < 0) { iCharLen = 1; }
			// Skip errors
			else              { ++iCharPos;   }
			iPos += iCharLen;

			if (iCharPos >= iMaxLen) { break; }
		}
		if (iCharPos >= iMaxLen)
		{
			sData = std::string(sData, 0, iPos);
			sData.append(aArguments[0].GetString());
		}

		oCDTRetVal = sData;
		return 0;
	}

return -1;
}

//
// Get function name
//
CCHAR_P FnMBTruncate::GetName() { return "mb_truncate"; }

//
// A destructor
//
FnMBTruncate::~FnMBTruncate() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnConcat
//

//
// Constructor
//
FnConcat::FnConcat()
{
	;;
}

//
// Handler
//
INT_32 FnConcat::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// Check number of parameters
	if (iArgNum == 0) { oCDTRetVal = ""; return 0; }

	if (iArgNum == 1) { oCDTRetVal = aArguments[0].GetString(); return 0; }

	std::string sResult;
	for(INT_32 iPos = iArgNum - 1; iPos >=0; --iPos) { sResult.append(aArguments[iPos].GetString()); }

	oCDTRetVal = sResult;

return 0;
}

//
// Get function name
//
CCHAR_P FnConcat::GetName() { return "concat"; }

//
// A destructor
//
FnConcat::~FnConcat() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnSubstring
//

//
// Constructor
//
FnSubstring::FnSubstring()
{
	;;
}

//
// Handler
//
INT_32 FnSubstring::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// Check number of parameters
	if (iArgNum == 0) { oCDTRetVal = ""; return 0; }

	// substr('foobar', 2) -> 'obar'
	if (iArgNum == 2)
	{
		const UINT_32      iOffset = UINT_32(aArguments[0].GetInt());
		const std::string  sResult = aArguments[1].GetString();

		if (sResult.size() < iOffset) { oCDTRetVal = ""; }
		else                          { oCDTRetVal = sResult.substr(iOffset); }

		return 0;
	}
	// substr('foobar', 2, 3) -> 'oba'
	if (iArgNum == 3)
	{
		const UINT_32      iBytes  = UINT_32(aArguments[0].GetInt());
		const UINT_32      iOffset = UINT_32(aArguments[1].GetInt());
		const std::string  sResult = aArguments[2].GetString();

		if (sResult.size() < iOffset) { oCDTRetVal = ""; }
		else                          { oCDTRetVal = sResult.substr(iOffset, iBytes); }

		return 0;
	}
	// substr('foobar', 2, 3, '1234567') -> 'fo1234567r'
	if (iArgNum == 4)
	{
		std::string        sReplacement = aArguments[0].GetString();
		const UINT_32      iBytes       = UINT_32(aArguments[1].GetInt());
		const UINT_32      iOffset      = UINT_32(aArguments[2].GetInt());
		const std::string  sTMP         = aArguments[3].GetString();

		if (iOffset > sTMP.size()) { oCDTRetVal = ""; return 0; }

		std::string  sResult(sTMP, 0, iOffset);
		sResult.append(sReplacement);
		if (iBytes + iOffset > sTMP.size()) { oCDTRetVal = sResult; return 0; }

		sResult.append(sTMP, iBytes + iOffset, std::string::npos);
		oCDTRetVal = sResult;
		return 0;
	}

return -1;
}

//
// Get function name
//
CCHAR_P FnSubstring::GetName() { return "substr"; }

//
// A destructor
//
FnSubstring::~FnSubstring() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnMBSubstring
//

//
// Constructor
//
FnMBSubstring::FnMBSubstring()
{
	;;
}

//
// Handler
//
INT_32 FnMBSubstring::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// Check number of parameters
	if (iArgNum == 0) { oCDTRetVal = ""; return 0; }

	// substr('foobar', 2) -> 'obar'
	if (iArgNum == 2)
	{
		const UINT_32      iOffset = UINT_32(aArguments[0].GetInt());
		const std::string  sResult = aArguments[1].GetString();

		CCHAR_P  szStart  = sResult.data();
		CCHAR_P  szEnd    = sResult.data() + sResult.size();
		INT_32   iPos     = 0;
		UINT_32  iCharPos = 0;
		for(;;)
		{
			INT_32 iCharLen = utf_charlen(szStart + iPos, szEnd);

			if (iCharLen == -3) { break; }

			// Check character length
			if (iCharLen < 0) { iCharLen = 1; }
			// Skip errors
			else              { ++iCharPos;   }
			iPos += iCharLen;

			if (iCharPos >= iOffset) { break; }
		}

		if (iCharPos < iOffset) { oCDTRetVal = ""; }
		else                    { oCDTRetVal = sResult.substr(iPos); }

		return 0;
	}
	// substr('foobar', 2, 3) -> 'oba'
	if (iArgNum == 3)
	{
		const UINT_32      iBytes  = UINT_32(aArguments[0].GetInt());
		const UINT_32      iOffset = UINT_32(aArguments[1].GetInt());
		const std::string  sResult = aArguments[2].GetString();

		CCHAR_P  szStart  = sResult.data();
		CCHAR_P  szEnd    = sResult.data() + sResult.size();
		INT_32   iPos     = 0;
		UINT_32  iCharOffset = 0;
		UINT_32  iCharPos = 0;
		for(;;)
		{
			INT_32 iCharLen = utf_charlen(szStart + iPos, szEnd);

			if (iCharLen == -3) { break; }

			// Check character length
			if (iCharLen < 0) { iCharLen = 1; }
			// Skip errors
			else              { ++iCharPos;   }
			iPos += iCharLen;

			if (iCharPos == iOffset)          { iCharOffset = iPos; }

			if (iCharPos == iOffset + iBytes) { break; }
		}

		if (sResult.size() < iCharOffset) { oCDTRetVal = ""; }
		else                              { oCDTRetVal = sResult.substr(iCharOffset, iPos - iCharOffset); }

		return 0;
	}
	// substr('foobar', 2, 3, '1234567') -> 'fo1234567r'
	if (iArgNum == 4)
	{
		std::string        sReplacement = aArguments[0].GetString();
		const UINT_32      iBytes       = UINT_32(aArguments[1].GetInt());
		const UINT_32      iOffset      = UINT_32(aArguments[2].GetInt());
		const std::string  sTMP         = aArguments[3].GetString();

		CCHAR_P  szStart  = sTMP.data();
		CCHAR_P  szEnd    = sTMP.data() + sTMP.size();
		UINT_32  iPos     = 0;
		UINT_32  iCharOffset = 0;
		UINT_32  iCharPos = 0;
		for(;;)
		{
			INT_32 iCharLen = utf_charlen(szStart + iPos, szEnd);

			if (iCharLen == -3) { break; }

			// Check character length
			if (iCharLen < 0) { iCharLen = 1; }
			// Skip errors
			else              { ++iCharPos;   }
			iPos += iCharLen;

			if (iCharPos == iOffset)          { iCharOffset = iPos; }

			if (iCharPos == iOffset + iBytes) { break; }
		}
		if (sTMP.size() < iCharOffset) { oCDTRetVal = ""; return 0; }

		std::string  sResult(sTMP, 0, iCharOffset);
		sResult.append(sReplacement);

		if (iPos == sTMP.size()) { oCDTRetVal = sResult; return 0; }

		sResult.append(sTMP, iPos, std::string::npos);
		oCDTRetVal = sResult;
		return 0;
	}

return -1;
}

//
// Get function name
//
CCHAR_P FnMBSubstring::GetName() { return "mb_substr"; }

//
// A destructor
//
FnMBSubstring::~FnMBSubstring() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnListElement
//

//
// Constructor
//
FnListElement::FnListElement()
{
	;;
}

//
// Handler
//
INT_32 FnListElement::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum <= 1) { return -1; }

	// Element index
	const UINT_32 iLastElement = iArgNum - 1;

	// Element
	const UINT_32 iElement = aArguments[iLastElement].GetInt();

	// Element does not exist?
	if (iElement >= iLastElement) { return -1; }

	// Return element
	oCDTRetVal = aArguments[iLastElement - iElement - 1];

return 0;
}

//
// Get function name
//
CCHAR_P FnListElement::GetName() { return "list_element"; }

//
// A destructor
//
FnListElement::~FnListElement() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnArrayElement
//

//
// Constructor
//
FnArrayElement::FnArrayElement()
{
	;;
}

//
// Handler
//
INT_32 FnArrayElement::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 2) { return -1; }

	// Second argument *MUST* be an ARRAY
	if (aArguments[0].GetType() != CDT::ARRAY_VAL) { return -1; }

	// Element index
	const UINT_32 iElement = aArguments[1].GetInt();

	if (aArguments[0].Size() <= iElement)
	{
		oCDTRetVal = CDT();
		return 0;
	}

	// Return element
	oCDTRetVal = aArguments[0][iElement];

return 0;
}

//
// Get function name
//
CCHAR_P FnArrayElement::GetName() { return "array_element"; }

//
// A destructor
//
FnArrayElement::~FnArrayElement() throw() { ;; }

} // namespace CTPP
// End.
